def print_block(blocks, indent=0, prefix=""):
    indentation = "    " * indent
    prefix_indentation = prefix + indentation
    prefix_indent_length = len(prefix_indentation)

    lines_to_print = [""]
    lines_lengths = [0]

    for string in blocks:
        current_line_length = lines_lengths[-1]
        new_length = current_line_length + len(string)
        total_length = prefix_indent_length + new_length + len(" &")

        if total_length < 120:
            lines_to_print[-1] += string
            lines_lengths[-1] = new_length
        else:
            lines_to_print[-1] += " &"
            lines_to_print.append(string)
            lines_lengths.append(len(string))

    result = "\n".join(prefix_indentation + line for line in lines_to_print) + "\n"
    return result


def get_shape_blocks(shape, fortran_order=True):
    """
    Construct the code blocks for the shape of an array.
    If fortran_order is False, the shape will be reversed.
    E.g.
        - shape (slice(2, 4, None), 5)

            -> ["(", "2", ":", "4", ",", "5", ")"] # fortran_order=True
            -> ["(", "5", ",", "2", ":", "4", ")"] # fortran_order=False

    """
    from numeta.syntax.settings import settings

    lbound = settings.array_lower_bound

    result = ["("]

    def convert(element):
        if element is None:
            # Assumed size array
            return [str(lbound), ":", "*"]
        elif isinstance(element, int):
            return [str(lbound), ":", str(element + (lbound - 1))]
        elif isinstance(element, slice):
            shift_end = None
            if element.start is None:
                start = [str(lbound)]

                shift_end = lbound - 1

            elif isinstance(element.start, int):
                start = [str(element.start)]
            else:
                start = element.start.get_code_blocks()

            # Note if the start is None, it has not be setted so we have to shift the end

            if element.stop is None:
                stop = [""]
            elif isinstance(element.stop, int):
                if shift_end is not None:
                    stop = [str(element.stop + shift_end)]
                else:
                    stop = [str(element.stop)]
            else:
                if shift_end is not None:
                    stop = (element.stop + shift_end).get_code_blocks()
                else:
                    stop = element.stop.get_code_blocks()

            if element.step is not None:
                raise NotImplementedError("Step in array dimensions is not implemented yet")
            return start + [":"] + stop

        else:
            # probably a variable or an expression
            return [
                str(lbound),
                ":",
                *(element + (lbound - 1)).get_code_blocks(),
            ]

    if isinstance(shape, tuple):
        dims = [convert(d) for d in shape]

        if not fortran_order:
            dims = dims[::-1]

        result += dims[0]
        for dim in dims[1:]:
            result += [",", " "]
            result += dim
    else:
        result += convert(shape)

    result += [")"]

    return result


# def get_variables(element):
#
#    if hasattr(element, "get_variables"):
#        return element.get_variables()
#
#    elif isinstance(element, (tuple, list)):
#        variables = []
#
#        for e in element:
#            variables += get_variables(e)
#
#        return variables
#
#    elif isinstance(element, slice):
#        variables = []
#        variables += get_variables(element.start)
#        variables += get_variables(element.stop)
#        variables += get_variables(element.step)
#        return variables
#    else:
#        return []


def get_nested_dependencies_or_declarations(entities, curr_module, for_module=False):
    """
    This function takes a set of entities and returns the dependencies and declarations of the entities.
    If for_module is True, it will return the declarations of the entities that are in the same module as curr_module.
    Important: Reverse the order of the declarations to make sure that the dependencies are declared before the entities.
    """
    from numeta.syntax.module import builtins_module

    dependencies = set()
    declarations = {}

    # revese to preserve the order of the entities
    for entity in entities[::-1]:
        if entity.parent is builtins_module:
            # No need to declare
            continue
        if not for_module:
            if entity.parent is None:
                # It is a local variable, so we need to declare
                declarations[entity.name] = entity.get_declaration()
            elif entity.parent is curr_module:
                # It is in the current module
                continue
            else:
                dependencies.add((entity.parent, entity))
        else:
            if entity.parent is None or entity.parent is curr_module:
                declarations[entity.name] = entity.get_declaration()
            else:
                dependencies.add((entity.parent, entity))

    new_declarations = declarations.copy()
    while new_declarations:
        new_entities = []
        for declaration in new_declarations.values():
            for variable in declaration.extract_entities():
                if variable not in new_entities and variable not in entities:
                    new_entities.append(variable)

        entities.extend(new_entities)

        # Now we can add the dependencies or define the local variables
        new_declarations = {}
        for entity in new_entities:
            if entity.parent is builtins_module:
                continue
            if not for_module:
                if entity.parent is None:
                    if entity.name not in declarations:
                        new_declarations[entity.name] = entity.get_declaration()
                elif entity.parent is curr_module:
                    continue
                else:
                    dependencies.add((entity.parent, entity))
            else:
                if entity.parent is None or entity.parent is curr_module:
                    declarations[entity.name] = entity.get_declaration()
                else:
                    dependencies.add((entity.parent, entity))

        declarations.update(new_declarations)

    return dependencies, {k: v for k, v in reversed(list(declarations.items()))}


def divide_variables_and_derived_types(declarations):
    from .variable_declaration import VariableDeclaration
    from .derived_type_declaration import DerivedTypeDeclaration
    from .subroutine_declaration import SubroutineDeclaration

    variable_declarations = {}
    derived_type_declarations = {}
    subroutine_declaration = {}

    for name, declaration in declarations.items():
        if isinstance(declaration, VariableDeclaration):
            variable_declarations[name] = declaration
        elif isinstance(declaration, DerivedTypeDeclaration):
            derived_type_declarations[name] = declaration
        elif isinstance(declaration, SubroutineDeclaration):
            subroutine_declaration[name] = declaration
        else:
            raise NotImplementedError(f"Unknown declaration type: {declaration}")

    return variable_declarations, derived_type_declarations, subroutine_declaration



from .fortran_type import FortranType
from .variable import Variable
from .derived_type import DerivedType
from .subroutine import Subroutine
from .module import Module, ExternalModule
from .scope import Scope
from .settings import settings

from .expressions import LiteralNode

false = LiteralNode(False)
FALSE = false
true = LiteralNode(True)
TRUE = true

from .statements import Assignment

from .statements import PointerAssignment

from .statements import Call

CALL = Call
call = Call

from .statements import If

fif = If
IF = If

from .statements import ElseIf

else_if = ElseIf
ELSE_IF = ElseIf
elseif = ElseIf
ELSEIF = ElseIf

from .statements import Else

felse = Else
ELSE = Else

from .statements import SelectCase

select_case = SelectCase
SELECT_CASE = SelectCase

from .statements import Case

case = Case
CASE = Case

from .statements import Do

do = Do
DO = Do

from .statements import DoWhile

do_while = DoWhile
DO_WHILE = DoWhile

from .statements import Allocate

allocate = Allocate
ALLOCATE = Allocate

from .statements import Deallocate

deallocate = Deallocate
DEALLOCATE = Deallocate

from .statements import Return

freturn = Return
RETURN = Return

from .statements import Cycle

cycle = Cycle
CYCLE = Cycle

from .statements import Exit

fexit = Exit
EXIT = Exit

from .statements import Stop

stop = Stop
STOP = Stop

from .statements import Print

fprint = Print
PRINT = Print

from .statements import Comment

comment = Comment
COMMENT = Comment

from .expressions import Abs

abs = Abs
ABS = Abs

from .expressions import Neg

neg = Neg
NEG = Neg

from .expressions import Not

fnot = Not
NOT = Not

from .expressions import Allocated

allocated = Allocated
ALLOCATED = Allocated

from .expressions import Shape

shape = Shape
SHAPE = Shape

from .expressions import All

all = All
ALL = All

from .expressions import Real

real = Real
REAL = Real

from .expressions import Imag

imag = Imag
IMAG = Imag

from .expressions import Conjugate

conjugate = Conjugate
CONJUGATE = Conjugate

from .expressions import Complex

fcomplex = Complex
COMPLEX = Complex

from .expressions import Transpose

transpose = Transpose
TRANSPOSE = Transpose

from .expressions import Exp

exp = Exp
EXP = Exp

from .expressions import Sqrt

sqrt = Sqrt
SQRT = Sqrt

from .expressions import Floor

floor = Floor
FLOOR = Floor

from .expressions import Sin

sin = Sin
SIN = Sin

from .expressions import Cos

cos = Cos
COS = Cos

from .expressions import Tan

tan = Tan
TAN = Tan

from .expressions import Sinh

sinh = Sinh
SINH = Sinh

from .expressions import Cosh

cosh = Cosh
COSH = Cosh

from .expressions import Tanh

tanh = Tanh
TANH = Tanh

from .expressions import ASin

asin = ASin
ASIN = ASin

from .expressions import ACos

acos = ACos
ACOS = ACos

from .expressions import ATan

atan = ATan
ATAN = ATan

from .expressions import ATan2

atan2 = ATan2
ATAN2 = ATan2

from .expressions import Dotproduct

dot_product = Dotproduct
DOT_PRODUCT = Dotproduct

from .expressions import Rank

rank = Rank
RANK = Rank

from .expressions import Size

size = Size
SIZE = Size

from .expressions import Max

max = Max
MAX = Max

from .expressions import Maxval

maxval = Maxval
MAXVAL = Maxval

from .expressions import Min

min = Min
MIN = Min

from .expressions import Minval

minval = Minval
MINVAL = Minval

from .expressions import Iand

iand = Iand
IAND = Iand

from .expressions import Ior

ior = Ior
IOR = Ior

from .expressions import Xor

xor = Xor
XOR = Xor

from .expressions import Ishft

ishft = Ishft
ISHFT = Ishft

from .expressions import Ibset

ibset = Ibset
IBSET = Ibset

from .expressions import Ibclr

ibclr = Ibclr
IBCLR = Ibclr

from .expressions import Popcnt

popcnt = Popcnt
POPCNT = Popcnt

from .expressions import Trailz

trailz = Trailz
TRAILZ = Trailz

from .expressions import Sum

sum = Sum
SUM = Sum

from .expressions import Matmul

matmul = Matmul
MATMUL = Matmul



from abc import ABC, abstractmethod
from .base_node import Node


class NamedEntity(Node, ABC):
    """
    A named entity that can be referenced.

    Attributes
    ----------
    name : str
        The name of the entity.
    parent : None | Module | ExternalLibrary
        If the entity is local the parent is None, if it is a module variable it is a module, lastly if it ca be a library.

    Methods
    -------
    extract_entities():
        Extract the entity itself.
    get_code_blocks():
        Return the code blocks representing the entity.
    """

    def __init__(self, name, parent=None) -> None:
        self.name = name
        self.parent = parent

    def __hash__(self):
        return hash(self.name)

    @abstractmethod
    def get_declaration(self):
        pass

    def get_code_blocks(self):
        return [self.name]

    def extract_entities(self):
        yield self



from numeta.syntax import Variable, ExternalModule, FortranType
from numeta.array_shape import UNKNOWN


class IsoCBinding(ExternalModule):
    def __init__(self):
        super().__init__("iso_c_binding", None)

        self.c_int32 = Variable("c_int32_t", ftype=None)
        self.c_int64 = Variable("c_int64_t", ftype=None)
        self.c_size_t = Variable("c_size_t", ftype=None)
        self.c_float = Variable("c_float", ftype=None)
        self.c_double = Variable("c_double", ftype=None)
        self.c_float_complex = Variable("c_float_complex", ftype=None)
        self.c_double_complex = Variable("c_double_complex", ftype=None)
        self.c_bool = Variable("c_bool", ftype=None)
        self.c_char = Variable("c_char", ftype=None)
        self.c_ptr = Variable("c_ptr", ftype=None)

        self.add_variable(
            self.c_int32,
            self.c_int64,
            self.c_size_t,
            self.c_float,
            self.c_double,
            self.c_float_complex,
            self.c_double_complex,
            self.c_bool,
            self.c_char,
            self.c_ptr,
        )

        self.add_method(
            "c_f_pointer",
            [
                Variable("cptr", self.c_ptr),
                Variable("fptr", ftype=None),
            ],
            bind_c=False,
        )

        self.add_method(
            "c_loc",
            [
                Variable("x", ftype=None),
            ],
            result_=FortranType("type", self.c_ptr),
            bind_c=False,
        )


iso_c = IsoCBinding()

FInt32_c = FortranType("integer", iso_c.c_int32)
FInt64_c = FortranType("integer", iso_c.c_int64)
FSizet_c = FortranType("integer", iso_c.c_size_t)
FReal32_c = FortranType("real", iso_c.c_float)
FReal64_c = FortranType("real", iso_c.c_double)
FComplex32_c = FortranType("complex", iso_c.c_float_complex)
FComplex64_c = FortranType("complex", iso_c.c_double_complex)
FLogical_c = FortranType("logical", iso_c.c_bool)
FCharacter_c = FortranType("character", iso_c.c_char)
FPointer_c = FortranType("type", iso_c.c_ptr)



import numpy as np
import pytest
import numeta as nm


@pytest.mark.parametrize("n_args", range(1, 5))
@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
@pytest.mark.parametrize("shape", [(), (5,), (2, 3)])
def test_variable_number_of_arguments(n_args, dtype, shape):
    @nm.jit
    def fill(*args):
        for i, arg in enumerate(args):
            arg[:] = float(i)

    args = [np.empty(shape, dtype=dtype) for _ in range(n_args)]
    fill(*args)

    for i, arg in enumerate(args):
        expected = np.full(shape, float(i), dtype=dtype)
        np.testing.assert_allclose(arg, expected)


def test_no_args():
    @nm.jit
    def no_op(*args):
        pass

    no_op()



import numpy as np
from pathlib import Path
import tempfile
import importlib.util
import sys
import sysconfig
import pickle
import shutil

from numeta.external_library import ExternalLibrary

from .compiler import FortranCompiler
from .registry import register_compilation_target
from .settings import settings
from .builder_helper import BuilderHelper
from .syntax import Subroutine, Variable
from .datatype import size_t
from .capi_interface import CAPIInterface
from .array_shape import ArrayShape, SCALAR, UNKNOWN


from .signature import (
    convert_signature_to_argument_specs,
    get_signature_and_runtime_args,
    parse_function_parameters,
)


class NumetaCompilationTarget(ExternalLibrary):
    """
    To link NumetaFunctions when called by other NumetaFunctions
    """

    def __init__(
        self,
        name,
        symbolic_function,
        *,
        directory: None | Path = None,
        do_checks=False,
        compile_flags="-O3 -march=native",
    ):
        super().__init__(name, to_link=False)
        self.symbolic_function = symbolic_function
        if directory is None:
            directory = tempfile.mkdtemp()
        self.directory = Path(directory).absolute()
        self.directory.mkdir(exist_ok=True)
        self.do_checks = do_checks
        if isinstance(compile_flags, str):
            self.compile_flags = compile_flags.split()
        else:
            self.compile_flags = compile_flags

        self._obj_files = None
        self.compiled_with_capi_file = None
        self.capi_name = None

        # Register this instance for later inspection via the class registry
        register_compilation_target(self)

        # Object files of all the NumetaCompilationTarget needed by this one
        # Because to link (capi part) we need all the obj files
        self._nested_obj_files = None

        # Fortunately we dont need nested mod files to compile

    @property
    def obj_files(self):
        if self._obj_files is None:
            self._obj_files, self._include = self.compile_fortran()
        return [self._obj_files]

    @property
    def include(self):
        if self._obj_files is None:
            self._obj_files, self._include = self.compile_fortran()
        return [self._include]

    def copy(self, directory):
        result = NumetaCompilationTarget(
            self.name,
            self.symbolic_function,
            directory=directory,
            do_checks=self.do_checks,
            compile_flags=self.compile_flags,
        )

        new_dir = directory / self.name
        new_dir.mkdir(exist_ok=True)
        if self._obj_files is not None:
            new_obj_file = new_dir / self.obj_files[0].name
            shutil.copy(self.obj_files[0], new_obj_file)
            result._obj_files = [new_obj_file]
        if self.compiled_with_capi_file is not None:
            new_lib_file = new_dir / self.compiled_with_capi_file.name
            shutil.copy(self.compiled_with_capi_file, new_lib_file)
            result.compiled_with_capi_file = new_lib_file
            result.capi_name = self.capi_name
        if len(self.get_nested_obj_files()) != 0:
            raise NotImplementedError("Dumping of nested numeta calls not implemented yet")

        return result

    def get_nested_obj_files(self):
        obj_files = set()
        queue = self.symbolic_function.get_dependencies().values()
        while queue:
            new_queue = []
            for lib in queue:
                if isinstance(lib, NumetaCompilationTarget):
                    new_obj = lib.obj_files[0]
                    if new_obj not in obj_files:
                        obj_files.add(new_obj)
                        new_queue += lib.symbolic_function.get_dependencies().values()
            queue = new_queue
        return obj_files

    def compile_fortran(self):
        """
        Compile Fortran source files using gfortran and return the resulting object file.
        """
        compiler = FortranCompiler(self.compile_flags)
        return compiler.compile_fortran(
            name=self.name,
            directory=self.directory,
            source=self.symbolic_function.get_code(),
            dependencies=self.symbolic_function.get_dependencies().values(),
        )

    def compile_with_capi_interface(
        self,
        capi_name,
        capi_obj,
    ):
        """
        Compiles Fortran code and constructs a C API interface,
        then compiles them into a shared library and loads the module.

        Parameters:
            *args: Arguments to pass to compile_fortran_function.

        Returns:
            tuple: (compiled function, subroutine)
        """
        self.capi_name = capi_name

        self.compiled_with_capi_file = self.directory / f"lib{self.name}_module.so"

        libraries = [
            "gfortran",
            "mvec",  # link math vectorized version
            f"python{sys.version_info.major}.{sys.version_info.minor}",
        ]
        libraries_dirs = []
        include_dirs = [sysconfig.get_paths()["include"], np.get_include()]
        extra_objects = self.get_nested_obj_files()
        additional_flags = []

        for lib in self.symbolic_function.get_dependencies().values():

            if lib.obj_files is not None:
                extra_objects |= set(lib.obj_files)

            if lib.to_link:
                libraries.append(lib.name)
                if lib.directory is not None:
                    libraries_dirs.append(lib.directory)
                if lib.include is not None:
                    include_dirs.append(lib.include)
                if lib.additional_flags is not None:
                    if isinstance(lib.additional_flags, str):
                        additional_flags.extend(lib.additional_flags.split())
                    else:
                        additional_flags.append(lib.additional_flags)

        command = ["gcc"]
        command.extend(self.compile_flags)
        command.extend(["-fopenmp"])
        command.extend(["-fPIC", "-shared", "-o", str(self.compiled_with_capi_file)])
        command.extend([str(*self.obj_files), str(capi_obj)])
        command.extend([str(obj) for obj in extra_objects])
        command.extend([f"-l{lib}" for lib in libraries])
        command.extend([f"-L{lib_dir}" for lib_dir in libraries_dirs])
        command.extend([f"-I{inc_dir}" for inc_dir in include_dirs])
        command.extend(additional_flags)

        compiler = FortranCompiler(self.compile_flags)
        compiler.run_command(command, cwd=self.directory)

    def load_with_capi(self):
        if self.capi_name is None:
            raise ValueError("Function should be compiled before loading it")
        spec = importlib.util.spec_from_file_location(self.capi_name, self.compiled_with_capi_file)
        compiled_sub = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(compiled_sub)
        return getattr(compiled_sub, self.name)


class NumetaFunction:
    """
    Representation of a JIT-compiled function.
    """

    def __init__(
        self,
        func,
        directory=None,
        do_checks=True,
        compile_flags="-O3 -march=native",
        namer=None,
        inline: bool | int = False,
    ) -> None:
        super().__init__()
        self.name = func.__name__
        if directory is None:
            directory = tempfile.mkdtemp()
        self.directory = Path(directory).absolute()
        self.directory.mkdir(exist_ok=True)
        self.do_checks = do_checks
        self.compile_flags = compile_flags

        self.namer = namer
        self.inline = inline
        self._func = func

        # To store the dependencies of the compiled functions to other numeta generated functions.
        (
            self.params,
            self.fixed_param_indices,
            self.n_positional_or_default_args,
            self.catch_var_positional_name,
        ) = parse_function_parameters(func)

        # Variables to populate
        self._signature_to_name = {}
        self.return_signatures = {}  # Only needed if i create symbolic and after compile
        self._compiled_targets = {}
        self._fast_call = {}

    def dump(self, directory):
        """
        Dumps the compiled function to a file.
        """
        directory = Path(directory)
        directory.mkdir(exist_ok=True)

        # Copy the libraries to the new directory
        new_compiled_target = {}
        for signature, compiled_target in self._compiled_targets.items():
            new_compiled_target[signature] = compiled_target.copy(directory)

        with open(directory / f"{self.name}.pkl", "wb") as f:
            pickle.dump(self._signature_to_name, f)
            pickle.dump(self.return_signatures, f)
            pickle.dump(new_compiled_target, f)

    def load(self, directory):
        """
        Loads the compiled function from a file.
        """
        self._fast_call = {}
        with open(Path(directory) / f"{self.name}.pkl", "rb") as f:
            self._signature_to_name = pickle.load(f)
            self._return_signatures = pickle.load(f)
            self._compiled_targets = pickle.load(f)
        self.return_signatures = {}

    def get_name(self, signature):
        if signature not in self._signature_to_name:
            if self.namer is None:
                name = f"{self.name}_{len(self._signature_to_name)}"
            else:
                name = self.namer(*signature)
            self._signature_to_name[signature] = name
        return self._signature_to_name[signature]

    def get_signature_idx(self, func):
        return

    def get_symbolic_functions(self):
        return [v.symbolic_function for v in self._compiled_targets.values()]

    def run_symbolic(self, *args, **kwargs):
        return self._func(*args, **kwargs)

    def get_signature(self, *args, **kwargs):
        _, signature, _ = get_signature_and_runtime_args(
            args,
            kwargs,
            params=self.params,
            fixed_param_indices=self.fixed_param_indices,
            n_positional_or_default_args=self.n_positional_or_default_args,
            catch_var_positional_name=self.catch_var_positional_name,
        )
        return signature

    def __call__(self, *args, **kwargs):
        """
        **Note**: Support for fixed shapes for arrays works when not using numpy arrays.
        But if the function is called with numeta types hint like nm.float32[2, 3] it will create a symbolic function with fixed shapes arguments.
        Calling it is another story and not implemented yet.
        """

        builder = BuilderHelper.current_builder
        if builder is not None:
            # We are already contructing a symbolic function

            _, signature, runtime_args = get_signature_and_runtime_args(
                args,
                kwargs,
                params=self.params,
                fixed_param_indices=self.fixed_param_indices,
                n_positional_or_default_args=self.n_positional_or_default_args,
                catch_var_positional_name=self.catch_var_positional_name,
            )

            symbolic_fun = self.get_symbolic_function(signature)
            return_specs = self.return_signatures.get(signature, [])

            # first check the runtime arguments
            # they should be symbolic nodes
            from .syntax.tools import check_node

            runtime_args = [check_node(arg) for arg in runtime_args]

            # Optionally add the array descriptor for arrays with runtime-dependent dimensions
            full_runtime_args = []
            for arg in runtime_args:
                if settings.add_shape_descriptors and arg._shape.has_comptime_undefined_dims():
                    full_runtime_args.append(arg._get_shape_descriptor())
                full_runtime_args.append(arg)

            do_inline = False
            if isinstance(self.inline, bool):
                do_inline = self.inline
            elif isinstance(self.inline, int):
                if symbolic_fun.count_statements() <= self.inline:
                    do_inline = True

            return_arguments = []
            return_values = []
            return_pointers = []
            if return_specs:
                for dtype, rank in return_specs:
                    if rank == 0:
                        out_var = builder.generate_local_variables(
                            "fc_r",
                            ftype=dtype.get_fortran(),
                        )
                        return_arguments.append(out_var)
                        return_values.append(out_var)
                        continue

                    shape_var = builder.generate_local_variables(
                        "fc_out_shape",
                        ftype=size_t.get_fortran(bind_c=True),
                        shape=ArrayShape((rank,)),
                    )
                    return_arguments.append(shape_var)

                    array_shape = ArrayShape(tuple([None] * rank))

                    if settings.use_numpy_allocator:
                        from numeta.external_modules.iso_c_binding import FPointer_c, iso_c

                        out_ptr = builder.generate_local_variables(
                            "fc_out_ptr",
                            ftype=FPointer_c,
                        )
                        return_arguments.append(out_ptr)

                        out_array = builder.generate_local_variables(
                            "fc_r",
                            ftype=dtype.get_fortran(),
                            shape=array_shape,
                            pointer=True,
                        )
                        return_pointers.append((out_ptr, out_array, shape_var, rank))
                        return_values.append(out_array)
                    else:
                        out_array = builder.generate_local_variables(
                            "fc_r",
                            ftype=dtype.get_fortran(),
                            shape=array_shape,
                            allocatable=True,
                        )
                        return_arguments.append(out_array)
                        return_values.append(out_array)

            if do_inline:
                builder.inline(symbolic_fun, *full_runtime_args, *return_arguments)
            else:
                # This add a Call statement to the current builder
                symbolic_fun(*full_runtime_args, *return_arguments)

            for out_ptr, out_array, shape_var, rank in return_pointers:
                if rank == 1:
                    shape_fortran = shape_var
                else:
                    shape_fortran = shape_var[rank - 1 : 1 : -1]
                from numeta.external_modules.iso_c_binding import iso_c

                iso_c.c_f_pointer(out_ptr, out_array, shape_fortran)

            if return_specs:
                if len(return_values) == 1:
                    return return_values[0]
                return tuple(return_values)
        else:

            # TODO: probably overhead, to do in C?
            to_execute, signature, runtime_args = get_signature_and_runtime_args(
                args,
                kwargs,
                params=self.params,
                fixed_param_indices=self.fixed_param_indices,
                n_positional_or_default_args=self.n_positional_or_default_args,
                catch_var_positional_name=self.catch_var_positional_name,
            )

            if not to_execute:
                return self.get_symbolic_function(signature)

            return self.execute_function(signature, runtime_args)

    def get_symbolic_function(self, signature):
        if signature not in self._compiled_targets:
            self.construct_symbolic_function(signature)
        return self._compiled_targets[signature].symbolic_function

    def construct_symbolic_function(self, signature):
        name = self.get_name(signature)
        argument_specs = convert_signature_to_argument_specs(
            signature,
            params=self.params,
            fixed_param_indices=self.fixed_param_indices,
            n_positional_or_default_args=self.n_positional_or_default_args,
        )

        sub = Subroutine(name)
        builder = BuilderHelper(self, sub, signature)

        def convert_argument_spec_to_variable(arg_spec):
            """
            Converts an ArgumentSpec to a Variable.
            """
            ftype = arg_spec.datatype.get_fortran()
            if arg_spec.rank == 0:
                return Variable(arg_spec.name, ftype=ftype, shape=SCALAR, intent=arg_spec.intent)
            elif arg_spec.shape is UNKNOWN:
                return Variable(
                    arg_spec.name,
                    ftype=ftype,
                    shape=UNKNOWN,
                    intent=arg_spec.intent,
                )
            elif arg_spec.shape.has_comptime_undefined_dims():
                if settings.add_shape_descriptors:
                    # The shape will to be passed as a separate argument
                    dim_var = Variable(
                        f"shape_{arg_spec.name}",
                        ftype=size_t.get_fortran(bind_c=True),
                        shape=ArrayShape((arg_spec.rank,)),
                        intent="in",
                    )
                    sub.add_variable(dim_var)

                    shape = ArrayShape(
                        tuple([dim_var[i] for i in range(arg_spec.rank)]),
                        fortran_order=arg_spec.shape.fortran_order,
                    )
                else:
                    shape = UNKNOWN
                return Variable(
                    arg_spec.name,
                    ftype=ftype,
                    shape=shape,
                    intent=arg_spec.intent,
                )
            else:
                # The dimension is fixed
                return Variable(
                    arg_spec.name,
                    ftype=ftype,
                    shape=arg_spec.shape,
                    intent=arg_spec.intent,
                )

        symbolic_args = []
        symbolic_kwargs = {}
        for arg in argument_specs:
            if arg.is_comptime:
                if arg.is_keyword:
                    symbolic_kwargs[arg.name] = arg.comptime_value
                else:
                    symbolic_args.append(arg.comptime_value)
            else:
                var = convert_argument_spec_to_variable(arg)
                # Add the variable to the subroutine
                sub.add_variable(var)
                if arg.is_keyword:
                    symbolic_kwargs[arg.name] = var
                else:
                    symbolic_args.append(var)

        return_signature = builder.build(*symbolic_args, **symbolic_kwargs)
        self.return_signatures[signature] = return_signature

        self._compiled_targets[signature] = NumetaCompilationTarget(
            self.get_name(signature),
            sub,
            directory=self.directory / self.get_name(signature),
            do_checks=self.do_checks,
            compile_flags=self.compile_flags,
        )

        sub.parent = self._compiled_targets[signature]

    def compile_function(self, signature):
        if signature not in self._compiled_targets:
            self.construct_symbolic_function(signature)

        capi_name = f"{self.name}_capi"
        capi_interface = CAPIInterface(
            self.get_name(signature),
            module_name=capi_name,
            args_details=convert_signature_to_argument_specs(
                signature,
                params=self.params,
                fixed_param_indices=self.fixed_param_indices,
                n_positional_or_default_args=self.n_positional_or_default_args,
            ),
            return_specs=self.return_signatures[signature],
            directory=self.directory / self.get_name(signature),
            compile_flags=self.compile_flags,
            do_checks=self.do_checks,
        )
        capi_obj = capi_interface.generate()

        self._compiled_targets[signature].compile_with_capi_interface(capi_name, capi_obj)

    def load_function(self, signature):
        if signature not in self._compiled_targets:
            self.compile_function(signature)
        self._fast_call[signature] = self._compiled_targets[signature].load_with_capi()

    def execute_function(self, signature, runtime_args):
        if signature not in self._fast_call:
            self.load_function(signature)
        return self._fast_call[signature](*runtime_args)



import numeta as nm
import numpy as np


def test_omp():
    @nm.jit
    def test_mul(a, b, c):
        n_threads = nm.scalar(nm.i8, nm.omp.omp_get_max_threads())
        i_thread = nm.scalar(nm.i8, 0)

        for j in nm.prange(
            b.shape[1],
            shared=[a, b, c, a.shape[0].variable, b.shape[0].variable],
            schedule="static",
        ):
            for k in nm.range(b.shape[0]):
                for i in nm.range(a.shape[0]):
                    i_thread[:] = nm.omp.omp_get_thread_num()

                    nm.omp.atomic_update_add(c[i, j].real, a[i, k].real * b[k, j].real)
                    nm.omp.atomic_update_sub(c[i, j].real, a[i, k].imag * b[k, j].imag)
                    nm.omp.atomic_update_add(c[i, j].imag, a[i, k].real * b[k, j].imag)
                    nm.omp.atomic_update_add(c[i, j].imag, a[i, k].imag * b[k, j].real)

    n = 50

    a = np.random.rand(n, n) + 1j * np.random.rand(n, n)
    b = np.random.rand(n, n) + 1j * np.random.rand(n, n)

    c = np.zeros((n, n), dtype=np.complex128)

    test_mul(a, b, c)

    np.testing.assert_allclose(c, a.dot(b))


test_omp()



from numeta.syntax import SelectCase, Case


def cases(select, cases_range: range):
    if not isinstance(cases_range, range):
        raise ValueError("The second argument must be a range object")

    with SelectCase(select):
        for c in cases_range:
            with Case(c):
                yield c



import numeta as nm
import numpy as np
import ctypes.util
import pytest


def test_blas():
    if ctypes.util.find_library("blas") is None:
        pytest.skip("BLAS library not found")
    blas = nm.ExternalLibraryWrapper("blas")
    blas.add_method(
        "dgemm",
        [
            nm.char,
            nm.char,
            nm.i8,
            nm.i8,
            nm.i8,
            nm.f8,
            nm.f8[None],
            nm.i8,
            nm.f8[None],
            nm.i8,
            nm.f8,
            nm.f8[None],
            nm.i8,
        ],
        None,
        bind_c=False,
    )

    n = 100

    @nm.jit
    def matmul(a, b, c):
        blas.dgemm(
            "N",
            "N",
            b.shape[0],
            a.shape[1],
            c.shape[1],
            1.0,
            b,
            b.shape[0],
            a,
            a.shape[0],
            0.0,
            c,
            c.shape[0],
        )

    a = np.random.rand(n, n)
    b = np.random.rand(n, n)
    c = np.zeros((n, n))

    matmul(a, b, c)

    np.testing.assert_allclose(c, np.dot(a, b))



import numpy as np
import pytest
import numeta as nm


@pytest.mark.parametrize(
    "dtype",
    [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128],
)
def test_scalar_call(dtype):

    @nm.jit
    def fill(dtype: nm.comptime, a):
        nm_dtype = nm.get_datatype(dtype)

        a[0] = nm_dtype(50)

        tmp2 = nm_dtype(100)
        a[1] = tmp2

    a = np.empty(2, dtype=dtype)
    fill(dtype, a)

    expected = np.empty(2, dtype=dtype)
    expected[0] = 50.0
    expected[1] = 100.0

    np.testing.assert_allclose(a, expected)


@pytest.mark.parametrize(
    "dtype",
    [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128],
)
def test_array_call_matrix(dtype):

    @nm.jit
    def fill(a):
        nm_dtype = nm.get_datatype(dtype)
        a[:2, :2] = nm_dtype[2, 3](50)[:2, :2]

        tmp2 = nm_dtype[2](100)
        a[:, 2] = tmp2[:2]

    out = np.empty((2, 3), dtype=dtype)
    fill(out)

    expected = np.empty((2, 3), dtype=dtype)
    expected[:2, :2] = 50
    expected[:, 2] = 100

    np.testing.assert_allclose(out, expected)


@pytest.mark.parametrize(
    "dtype",
    [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128],
)
def test_array_call_3d(dtype):

    @nm.jit
    def fill(a):
        nm_dtype = nm.get_datatype(dtype)
        a[:] = nm_dtype[2, 2, 3](7)

    out = np.empty((2, 2, 3), dtype=dtype)
    fill(out)

    expected = np.full((2, 2, 3), 7, dtype=dtype)
    np.testing.assert_allclose(out, expected)


def test_variable_declaration_syntax():

    a = nm.float64(name="a")
    from numeta.syntax.statements import VariableDeclaration

    dec = VariableDeclaration(a)
    assert dec.print_lines() == ["real(c_double) :: a\n"]

    array = nm.int32[10, 20](name="array")
    dec = VariableDeclaration(array)
    assert dec.print_lines() == ["integer(c_int32_t), dimension(0:19, 0:9) :: array\n"]

    array_unknown = nm.int32[None](name="array_unknown")
    dec = VariableDeclaration(array_unknown)
    print(dec.print_lines())


test_variable_declaration_syntax()



import numpy as np
import numeta as nm


def test_cases():
    @nm.jit
    def cases(a) -> None:
        a[:] = 0.0
        for i in nm.range(n):
            for j in nm.cases(i, range(n)):
                a[j] = j

    n = 27
    a = np.empty(n, dtype=np.float64)

    cases(a)
    np.testing.assert_allclose(a, np.array(range(n), dtype=np.float64))



from .fortran_type import FortranType


class SyntaxSettings:
    def __init__(
        self,
        c_like: bool = False,
        int_kind: int = 8,
        float_kind: int = 8,
        complex_kind: int = 8,
        bool_kind: int = 1,
        char_kind: int = 1,
        order: str = "F",
    ):
        self.__subroutine_bind_c = False
        self.__derived_type_bind_c = False
        self.__array_lower_bound = 1
        self.__c_like_bounds = False
        self.__force_value = False
        self.order = order
        self.c_like = c_like

        if self.c_like:
            self.set_c_like()
        else:
            self.unset_c_like()

        self.set_default_integer_kind(int_kind)
        self.set_default_real_kind(float_kind)
        self.set_default_complex_kind(complex_kind)
        self.set_default_logical_kind(bool_kind)
        self.set_default_character_kind(char_kind)

    def set_c_like(self):
        self.c_like = True
        self.set_array_lower_bound(0)
        self.set_subroutine_bind_c()
        self.set_derived_type_bind_c()
        self.set_force_value()
        self.set_c_like_bounds()
        self.order = "C"

    def unset_c_like(self):
        self.c_like = False
        self.set_array_lower_bound(1)
        self.unset_subroutine_bind_c()
        self.unset_derived_type_bind_c()
        self.unset_force_value()
        self.unset_c_like_bounds()
        self.order = "F"

    def set_array_order(self, order: str):
        if order == "C":
            self.order = "C"
        elif order == "F":
            self.order = "F"
        else:
            raise ValueError(f"Order {order} not supported")

    # --- Direct FortranType setters --------------------------------------
    def set_default_fortran_type(self, ftype: FortranType):
        if not isinstance(ftype, FortranType):
            raise TypeError("ftype must be a FortranType")
        ftype_name = ftype.type
        if ftype_name == "integer":
            self.set_default_integer_kind(ftype.kind)
        elif ftype_name == "real":
            self.set_default_real_kind(ftype.kind)
        elif ftype_name == "complex":
            self.set_default_complex_kind(ftype.kind)
        elif ftype_name == "logical":
            self.set_default_logical_kind(ftype.kind)
        elif ftype_name == "character":
            self.set_default_character_kind(ftype.kind)
        else:
            raise NotImplementedError(f"Unsupported Fortran type {ftype_name}")

    # --- Kind setters ----------------------------------------------------
    def set_default_integer_kind(self, kind):
        self.DEFAULT_INTEGER_KIND = kind
        self.DEFAULT_INTEGER = FortranType("integer", kind)

    def set_default_real_kind(self, kind):
        self.DEFAULT_REAL_KIND = kind
        self.DEFAULT_REAL = FortranType("real", kind)

    def set_default_complex_kind(self, kind):
        self.DEFAULT_COMPLEX_KIND = kind
        self.DEFAULT_COMPLEX = FortranType("complex", kind)

    def set_default_logical_kind(self, kind):
        self.DEFAULT_LOGICAL_KIND = kind
        self.DEFAULT_LOGICAL = FortranType("logical", kind)

    def set_default_character_kind(self, kind):
        self.DEFAULT_CHARACTER_KIND = kind
        self.DEFAULT_CHARACTER = FortranType("character", kind)

    # --- Properties ------------------------------------------------------
    @property
    def subroutine_bind_c(self):
        return self.__subroutine_bind_c

    def set_subroutine_bind_c(self):
        self.__subroutine_bind_c = True

    def unset_subroutine_bind_c(self):
        self.__subroutine_bind_c = False

    @property
    def derived_type_bind_c(self):
        return self.__derived_type_bind_c

    def set_derived_type_bind_c(self):
        self.__derived_type_bind_c = True

    def unset_derived_type_bind_c(self):
        self.__derived_type_bind_c = False

    @property
    def array_lower_bound(self):
        return self.__array_lower_bound

    def set_array_lower_bound(self, value):
        try:
            self.__array_lower_bound = int(value)
        except ValueError as e:
            raise ValueError(f"Array lower bound must be an integer, got {value}") from e

    @property
    def c_like_bounds(self):
        return self.__c_like_bounds

    def set_c_like_bounds(self):
        self.__c_like_bounds = True

    def unset_c_like_bounds(self):
        self.__c_like_bounds = False

    @property
    def force_value(self):
        return self.__force_value

    def set_force_value(self):
        self.__force_value = True

    def unset_force_value(self):
        self.__force_value = False


settings = SyntaxSettings()



import numpy as np
from dataclasses import dataclass
from .external_modules.iso_c_binding import iso_c
from .syntax import FortranType, DerivedType
from .array_shape import ArrayShape, SCALAR, UNKNOWN


class DataTypeMeta(type):
    """Metaclass used for all data type classes."""

    _np_dtype = {}
    _ftype = {}
    _ftype_bind_c = {}

    def __new__(mcls, name, bases, attrs):
        cls = super().__new__(mcls, name, bases, attrs)

        # Register built-in numpy mapping when available
        np_type = attrs.get("_np_type")
        if np_type is not None:
            if np_type in DataTypeMeta._np_dtype:
                raise ValueError(f"DataType {np_type} already exists")
            DataTypeMeta._np_dtype[np_type] = cls

        ftype = attrs.get("_fortran_type")
        if ftype is not None:
            if ftype in DataTypeMeta._ftype:
                raise ValueError(f"FortranType {ftype} already exists")
            DataTypeMeta._ftype[ftype] = cls

        fortran_bind_c_type = attrs.get("_fortran_bind_c_type")
        if fortran_bind_c_type is not None:
            if fortran_bind_c_type in DataTypeMeta._ftype_bind_c:
                raise ValueError(f"FortranType {fortran_bind_c_type} already exists")
            DataTypeMeta._ftype_bind_c[fortran_bind_c_type] = cls

        return cls

    def __repr__(cls):
        return f"numeta.{cls._name}"

    def __call__(cls, *args, **kwargs):
        # StructType overrides this behaviour and must be instantiated normally
        value = args[0] if args else kwargs.get("value", None)
        name = kwargs.get("name", None)
        from .wrappers.scalar import scalar

        return scalar(cls, value, name=name)


class DataType(metaclass=DataTypeMeta):
    """Base class for all data type definitions."""

    _np_type = None
    _fortran_type = None
    _fortran_bind_c_type = None
    _cnp_type = None
    _capi_cast = staticmethod(lambda x: x)
    _name = "datatype"
    _is_struct = False
    _can_be_value = True

    @property
    def name(cls):
        return cls._name

    @classmethod
    def __class_getitem__(cls, key):
        if key is None:
            # It is a pointer
            return ArrayType(dtype=cls, shape=UNKNOWN)
        if not isinstance(key, tuple):
            key = (key,)
        new_key = []
        for k in key:
            if isinstance(k, slice):
                if k.start is not None or k.stop is not None or k.step is not None:
                    raise TypeError(f"Invalid type for array dimension: {type(k)}")
            elif not isinstance(k, int):
                raise TypeError(f"Invalid type for array dimension: {type(k)}")
            if isinstance(k, int) and k < 0:
                raise ValueError("Negative dimensions are not allowed")
            new_key.append(k if isinstance(k, int) else None)
        return ArrayType(dtype=cls, shape=ArrayShape(tuple(new_key)))

    @classmethod
    def is_np_dtype(cls, dtype):
        return dtype in DataTypeMeta._np_dtype

    @classmethod
    def from_np_dtype(cls, dtype):
        """Get the DataType class from a numpy dtype."""
        return DataTypeMeta._np_dtype[dtype]

    @classmethod
    def from_ftype(cls, ftype):
        """Get the DataType class from a FortranType."""
        if ftype in DataTypeMeta._ftype_bind_c:
            return DataTypeMeta._ftype_bind_c[ftype]
        return DataTypeMeta._ftype[ftype]

    @classmethod
    def is_struct(cls):
        return cls._is_struct

    @classmethod
    def can_be_value(cls):
        return cls._can_be_value

    @classmethod
    def get_numpy(cls):
        return cls._np_type

    @classmethod
    def get_cnumpy(cls):
        return cls._cnp_type

    @classmethod
    def get_fortran(cls, bind_c=None):
        if bind_c is None:
            from .settings import settings

            return cls._fortran_bind_c_type if settings.iso_C else cls._fortran_type
        return cls._fortran_bind_c_type if bind_c else cls._fortran_type

    @classmethod
    def get_capi_cast(cls, obj):
        return cls._capi_cast(obj)

    @classmethod
    def get_nbytes(cls):
        """Get the number of bytes used by this data type."""
        if cls._np_type is not None:
            return np.dtype(cls._np_type).itemsize
        raise NotImplementedError(f"DataType {cls._name} does not have a defined size")


@dataclass(frozen=True)
class ArrayType:
    """Helper object returned by DataType[x] to describe array types."""

    dtype: DataType
    shape: ArrayShape

    @classmethod
    def __class_getitem__(cls, key):
        """
        Create an ArrayType with a given order and shape.
        If key is 'C' is C-style order, if 'F' is Fortran-style order.
        """
        if key != "C" and key != "F":
            raise ValueError(f"Invalid order: {key}, must be 'C' or 'F'")
        return cls(dtype=cls.dtype, shape=ArrayShape(cls.shape.dims, fortran_order=(key == "F")))

    def __repr__(self):
        if self.shape is UNKNOWN:
            return f"{self.dtype._name}[*]"
        dims = ",".join(
            ":" if isinstance(d, slice) and d == slice(None) else str(d) for d in self.shape.dims
        )
        return f"{self.dtype._name}[{dims}]"

    def __call__(self, *args, **kwargs):
        value = args[0] if args else kwargs.get("value", None)
        name = kwargs.get("name", None)

        from .wrappers.empty import empty

        array = empty(self.shape, dtype=self.dtype, order=kwargs.get("order", "C"), name=name)
        if value is not None:
            array[:] = value

        return array


class int32(DataType):
    _np_type = np.int32
    _fortran_type = FortranType("integer", 4)
    _fortran_bind_c_type = FortranType("integer", iso_c.c_int32)
    _cnp_type = "npy_int32"
    _capi_cast = staticmethod(lambda x: f"PyLong_AsLongLong({x})")
    _name = "int32"


class int64(DataType):
    _np_type = np.int64
    _fortran_type = FortranType("integer", 8)
    _fortran_bind_c_type = FortranType("integer", iso_c.c_int64)
    _cnp_type = "npy_int64"
    _capi_cast = staticmethod(lambda x: f"PyLong_AsLongLong({x})")
    _name = "int64"


class size_t(DataType):
    _np_type = None
    _fortran_type = None
    _fortran_bind_c_type = FortranType("integer", iso_c.c_size_t)
    _cnp_type = "npy_intp"
    _capi_cast = staticmethod(lambda x: f"PyLong_AsLongLong({x})")
    _name = "size_t"


class float32(DataType):
    _np_type = np.float32
    _fortran_type = FortranType("real", 4)
    _fortran_bind_c_type = FortranType("real", iso_c.c_float)
    _cnp_type = "npy_float32"
    _capi_cast = staticmethod(lambda x: f"PyFloat_AsDouble({x})")
    _name = "float32"


class float64(DataType):
    _np_type = np.float64
    _fortran_type = FortranType("real", 8)
    _fortran_bind_c_type = FortranType("real", iso_c.c_double)
    _cnp_type = "npy_float64"
    _capi_cast = staticmethod(lambda x: f"PyFloat_AsDouble({x})")
    _name = "float64"


class complex64(DataType):
    _np_type = np.complex64
    _fortran_type = FortranType("complex", 4)
    _fortran_bind_c_type = FortranType("complex", iso_c.c_float_complex)
    _cnp_type = "npy_complex64"
    _capi_cast = staticmethod(
        lambda x: f"CMPLX(PyComplex_RealAsDouble({x}), PyComplex_ImagAsDouble({x}))",
    )
    _name = "complex64"


class complex128(DataType):
    _np_type = np.complex128
    _fortran_type = FortranType("complex", 8)
    _fortran_bind_c_type = FortranType("complex", iso_c.c_double_complex)
    _cnp_type = "npy_complex128"
    _capi_cast = staticmethod(
        lambda x: f"CMPLX(PyComplex_RealAsDouble({x}), PyComplex_ImagAsDouble({x}))"
    )
    _name = "complex128"


class bool8(DataType):
    _np_type = np.bool_
    _fortran_type = FortranType("logical", 1)
    _fortran_bind_c_type = FortranType("logical", iso_c.c_bool)
    _cnp_type = "npy_bool"
    _capi_cast = staticmethod(lambda x: f"PyObject_IsTrue({x})")
    _name = "bool8"


class char(DataType):
    _np_type = np.str_
    _fortran_type = FortranType("character", 1)
    _fortran_bind_c_type = FortranType("character", iso_c.c_char)
    _cnp_type = "npy_str"
    _capi_cast = staticmethod(lambda x: f"PyUnicode_AsUTF8({x})")
    _name = "char"


class StructType(DataType, metaclass=DataTypeMeta):
    """Metaclass used to build struct datatype classes."""

    _counter = 0
    _fortran_type = None
    _fortran_bind_c_type = None
    _np_type = None
    _cnp_type = None
    _capi_cast = staticmethod(lambda x: x)
    _name = "datatype"
    _is_struct = False
    _can_be_value = True
    _members = []

    @classmethod
    def c_declaration(cls):
        members_str = []
        for mname, dt, shape in cls._members:
            dec = f"{dt.get_cnumpy()} {mname}"
            if shape is not SCALAR and shape is not UNKNOWN:
                dec += "".join(f"[{d}]" for d in shape.dims)
            members_str.append(dec)
        members_join = "; ".join(members_str)
        return f"typedef struct {{ {members_join} ;}} {cls._cnp_type};\n"

    @classmethod
    def __class_getitem__(cls, key) -> ArrayType:
        if key is None:
            # It is a pointer
            return ArrayType(dtype=cls, shape=UNKNOWN)
        if not isinstance(key, tuple):
            key = (key,)
        return ArrayType(dtype=cls, shape=ArrayShape(key))


def make_struct_type(np_dtype, members, name=None):
    """Create (or retrieve) a struct datatype class for ``members``."""

    key = tuple(members)
    if key in DataTypeMeta._np_dtype:
        return DataTypeMeta._np_dtype[key]

    if name is None:
        name = f"struct{StructType._counter}"

    StructType._counter += 1

    fortran_type = FortranType(
        "type",
        DerivedType(
            name,
            [(mname, dt.get_fortran(), dim) for mname, dt, dim in members],
        ),
    )

    attrs = {
        "_name": name,
        "name": name,
        "_members": members,
        "members": members,
        "_np_type": np_dtype,
        "_cnp_type": name,
        "_fortran_type": fortran_type,
        "_fortran_bind_c_type": fortran_type,
        "_is_struct": True,
        "_can_be_value": False,
    }

    new_cls = type(name, (StructType,), attrs)

    return new_cls


def get_struct_from_np_dtype(np_dtype):
    if np_dtype in DataTypeMeta._np_dtype:
        return DataTypeMeta._np_dtype[np_dtype]

    fields = []
    for name, (np_d, _) in np_dtype.base.fields.items():
        if np_d in DataTypeMeta._np_dtype:
            dtype = DataTypeMeta._np_dtype[np_d]
        elif np_d.base.type in DataTypeMeta._np_dtype:
            dtype = DataTypeMeta._np_dtype[np_d.base.type]
        elif np_d.fields is not None or (np_d.base.fields is not None):
            dtype = get_struct_from_np_dtype(np_d)
        else:
            raise ValueError(f"Invalid dtype {np_d.base.type}, {np_d.fields}")

        shape = SCALAR if len(np_d.shape) == 0 else ArrayShape(np_d.shape)
        fields.append((name, dtype, shape))

    struct_cls = make_struct_type(np_dtype, fields)
    DataTypeMeta._np_dtype[np_dtype] = struct_cls
    return struct_cls


def get_datatype(dtype):
    #
    # Numeta DataType
    #
    if isinstance(dtype, type) and issubclass(dtype, DataType):
        return dtype
    #
    # Python numeric types
    #
    if dtype is int:
        return int64
    elif dtype is float:
        return float64
    elif dtype is complex:
        return complex128
    elif dtype is bool:
        return bool8

    #
    # Numpy dtypes
    #

    # Canonalize dtype
    if isinstance(dtype, np.dtype):
        base = dtype.base.type
    else:
        base = getattr(dtype, "base", dtype).type if hasattr(dtype, "type") else dtype
        if isinstance(base, np.dtype):
            base = base.type

    # Check if it is a numpy dtype
    if DataType.is_np_dtype(base):
        return DataType.from_np_dtype(base)

    # It could be a struct
    if hasattr(dtype, "fields"):
        return get_struct_from_np_dtype(dtype)

    raise ValueError(f"Invalid dtype {dtype}")



from .statement import Statement
from numeta.syntax.scope import Scope
from numeta.syntax.tools import check_node


class Call(Statement):
    def __init__(
        self,
        function,
        *arguments,
        add_to_scope=True,
    ):
        super().__init__(add_to_scope=add_to_scope)
        self.function = function
        self.arguments = [check_node(arg) for arg in arguments]

    @property
    def children(self):
        """Return the child nodes of the call."""
        return [self.function] + self.arguments

    def get_with_updated_variables(self, variables_couples):
        new_arguments = [
            arg.get_with_updated_variables(variables_couples) for arg in self.arguments
        ]
        return Call(self.function, *new_arguments, add_to_scope=False)

    def get_code_blocks(self):
        if isinstance(self.function, str):
            result = ["call", " ", self.function]
        else:
            result = ["call", " ", self.function.name]

        result += ["("]
        for arg in self.arguments:
            result += arg.get_code_blocks()
            result += [", "]
        if result[-1] == ", ":
            result.pop()
        result += [")"]

        return result



from .nodes import NamedEntity
from .expressions import ExpressionNode
from numeta.array_shape import ArrayShape, SCALAR


class Variable(NamedEntity, ExpressionNode):
    def __init__(
        self,
        name,
        ftype,
        shape=SCALAR,
        intent=None,
        pointer=False,
        target=False,
        allocatable=False,
        parameter=False,
        assign=None,
        parent=None,
        bind_c=False,
    ):
        super().__init__(name, parent=parent)
        self.__ftype = ftype
        if not isinstance(shape, ArrayShape):
            self.__shape = ArrayShape(shape, fortran_order=True)
        else:
            self.__shape = shape
        self.allocatable = allocatable
        self.parameter = parameter
        self.assign = assign
        self.intent = intent
        self.pointer = pointer
        self.target = target
        # Note that bind c make the variable global
        self.bind_c = bind_c

        from .module import Module

        if isinstance(self.parent, Module):
            self.parent.add_variable(self)

    @property
    def _ftype(self):
        return self.__ftype

    @property
    def _shape(self):
        return self.__shape

    def _set_shape(self, shape):
        if not isinstance(shape, ArrayShape):
            self.__shape = ArrayShape(shape, fortran_order=self._shape.fortran_order)
        else:
            self.__shape = shape

    def get_with_updated_variables(self, variables_couples):
        for old_variable, new_variable in variables_couples:
            if old_variable.name == self.name:
                return new_variable
        return self

    def get_declaration(self):
        from .statements import VariableDeclaration

        return VariableDeclaration(self)

    @property
    def real(self):
        from .expressions import Re

        return Re(self)

    @real.setter
    def real(self, value):
        from .expressions import Re
        from .statements import Assignment

        return Assignment(Re(self), value)

    @property
    def imag(self):
        from .expressions import Im

        return Im(self)

    @imag.setter
    def imag(self, value):
        from .expressions import Im
        from .statements import Assignment

        return Assignment(Im(self), value)

    @property
    def shape(self):
        return self._shape.dims

    def __setitem__(self, key, value):
        """Does nothing, but allows to use variable[key] = value"""
        from .statements import Assignment

        if isinstance(key, slice) and key.start is None and key.stop is None and key.step is None:
            # if the variable is assigned to itself, do nothing, needed for the += and -= operators
            if self is value:
                return
            Assignment(self, value)
        else:
            Assignment(self[key], value)

    def __ilshift__(self, other):
        from .statements import Assignment

        Assignment(self, other)
        return self

    def __iadd__(self, other):
        from .statements import Assignment

        Assignment(self, self + other)
        # need to return same, no real assignment
        return self

    def __isub__(self, other):
        from .statements import Assignment

        Assignment(self, self - other)
        # need to return same, no real assignment
        return self

    def copy(self):
        return Variable(
            self.name,
            self._ftype,
            shape=self._shape,
            intent=self.intent,
            pointer=self.pointer,
            target=self.target,
            allocatable=self.allocatable,
            parameter=self.parameter,
            assign=self.assign,
            parent=self.parent,
        )



import numpy as np

from numeta.builder_helper import BuilderHelper
from numeta.datatype import DataType, FortranType, get_datatype
from numeta.external_modules.iso_c_binding import iso_c


def cast(variable, dtype: DataType | FortranType | np.generic):

    if isinstance(dtype, FortranType):
        ftype = dtype
    else:
        ftype = get_datatype(dtype).get_fortran()

    builder = BuilderHelper.get_current_builder()
    pointer = builder.generate_local_variables(
        "fc_v",
        ftype=ftype,
        pointer=True,
    )

    variable.target = True
    iso_c.c_f_pointer(iso_c.c_loc(variable), pointer)

    return pointer



import numeta as nm
import numpy as np
import pytest


def test_nan():

    @nm.jit
    def check_nan(a, is_nan):
        is_nan[0] = a[0] != a[0]

    a = np.array((1,), dtype=np.float64)
    a[0] = np.nan

    is_nan = np.zeros((1,), dtype=np.bool_)

    check_nan(a, is_nan)

    np.testing.assert_equal(is_nan[0], a[0] != a[0])



class ExternalLibrary:
    """
    A class to represent an external library.
    It is used to link external libraries to the fortran code.
    Is is child of Module class, where the module is hidden.
    Can contain ExternalModule objects.
    """

    def __init__(
        self,
        name,
        directory=None,
        include=None,
        obj_files=None,
        additional_flags=None,
        to_link=True,
    ):
        """
        Directory is the path to the directory where the external library to link is located.
        Include is the path of the header file to include.
        """
        self.name = name
        self.hidden = True
        self.external = True
        self.directory = directory
        self._include = include
        self._obj_files = obj_files
        self.additional_flags = additional_flags
        self.to_link = to_link

        self.modules = {}
        self.subroutines = {}
        self.variables = {}

    @property
    def obj_files(self):
        return self._obj_files

    @property
    def include(self):
        return self._include



import numpy as np
import numeta as nm


def test_call_array_scalar():

    @nm.jit
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a)

    a = np.zeros((), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((), dtype=np.int64)
    expected[...] = 1
    np.testing.assert_equal(a, expected)


def test_call_array():

    @nm.jit
    def callee(n, a):
        a[:, 2] = n

    @nm.jit
    def caller(n, a):
        callee(n, a)

    a = np.zeros((5, 10), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[:, 2] = 1
    np.testing.assert_equal(a, expected)


def test_call_getitem_scalar():

    @nm.jit
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a[3, 7])

    a = np.zeros((5, 10), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[3, 7] = 1
    np.testing.assert_equal(a, expected)


def test_call_getitem_slice():

    @nm.jit
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(50, a[:])
        callee(n, a[:2, :3])
        callee(n + 1, a[3:4, 8:])
        callee(n + 2, a[:, 7])

    a = np.zeros((5, 10), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[:] = 50
    expected[:2, :3] = 1
    expected[3:4, 8:] = 2
    expected[:, 7] = 3
    np.testing.assert_equal(a, expected)


def test_call_getitem_slice_runtime_dep():

    @nm.jit
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, m, a):
        callee(50, a[:])
        callee(n, a[:n, : m // 2])
        callee(2, a[2 : n + 1, m - 2 :])
        callee(3, a[:, n])
        callee(1, a[:, n : m - n])

    n = 4
    m = 9
    a = np.zeros((5, 10), dtype=np.int64)
    caller(n, m, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[:] = 50
    expected[:n, : m // 2] = n
    expected[2 : n + 1, m - 2 :] = 2
    expected[:, n] = 3
    expected[:, n : m - n] = 1
    np.testing.assert_equal(a, expected)


def test_call_getattr_scalar():

    @nm.jit
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a["x"])

    dtype = np.dtype([("x", np.int64), ("y", np.float64, (2, 2))], align=True)

    a = np.zeros((), dtype=dtype)
    caller(1, a)

    expected = np.zeros((), dtype=dtype)
    expected["x"] = 1
    np.testing.assert_equal(a, expected)


def test_call_getattr_array():

    @nm.jit
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a["y"])

    dtype = np.dtype([("x", np.int64), ("y", np.float64, (2, 2))], align=True)

    a = np.zeros((), dtype=dtype)
    caller(2.0, a)

    expected = np.zeros((), dtype=dtype)
    expected["y"] = 2.0
    np.testing.assert_equal(a, expected)


def test_call_matmul():

    @nm.jit
    def callee(a, d):
        a[:] = d

    @nm.jit
    def caller(a, b, c):
        callee(c, nm.matmul(b, a))

    n = 10
    a = np.random.random((n, n))
    b = np.random.random((n, n))
    c = np.zeros((10, 10))
    caller(a, b, c)

    expected = a @ b
    np.testing.assert_equal(c, expected)


def test_call_matmul_fortran_order():

    @nm.jit
    def callee(a, d):
        a[:] = d

    @nm.jit
    def caller(a, b, c):
        callee(c, nm.matmul(a, b))

    n = 10
    a = np.random.random((n, n)).astype(np.float64, order="F")
    b = np.random.random((n, n)).astype(np.float64, order="F")
    c = np.zeros((10, 10), order="F")
    caller(a, b, c)

    expected = a @ b
    np.testing.assert_equal(c, expected)


def test_nested_calls():
    @nm.jit
    def inner(n, arr):
        arr[0] = n

    @nm.jit
    def middle(n, arr):
        inner(n, arr)
        arr[1] = n + 1

    @nm.jit
    def caller(n, arr):
        middle(n, arr)

    arr = np.zeros(2, dtype=np.int64)
    caller(5, arr)

    expected = np.array([5, 6], dtype=np.int64)
    np.testing.assert_equal(arr, expected)


def test_nested_calls_with_literals():
    @nm.jit
    def inner(n, arr):
        arr[0] = n

    @nm.jit
    def middle(n, arr):
        inner(2, arr)
        arr[1] = n + 1

    @nm.jit
    def caller(arr):
        middle(3, arr)

    arr = np.zeros(2, dtype=np.int64)
    caller(arr)

    expected = np.array([2, 4], dtype=np.int64)
    np.testing.assert_equal(arr, expected)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .numeta_function import NumetaCompilationTarget

_registry: list["NumetaCompilationTarget"] = []


def register_compilation_target(target: "NumetaCompilationTarget") -> None:
    _registry.append(target)


def registered_functions() -> list["NumetaCompilationTarget"]:
    """Return a snapshot of all instantiated :class:`NumetaFunction` objects."""
    return list(_registry)


def clear_registered_functions() -> None:
    """Clear the registry of instantiated :class:`NumetaFunction` objects."""
    _registry.clear()



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_CompTime(dtype):
    n = 100

    @nm.jit
    def mul(ct: nm.comptime, a, b, c):
        for i in nm.range(a.shape[0]):
            for k in nm.range(b.shape[0]):
                c[i, :] += a[i, k] * b[k, :]

    a = np.random.rand(n, n).astype(dtype)
    b = np.random.rand(n, n).astype(dtype)
    c = np.zeros((n, n), dtype=dtype)

    mul(1.0, a, b, c)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(c, np.dot(a, b), atol=0)
    else:
        np.testing.assert_allclose(c, np.dot(a, b), rtol=float(10e2 * np.finfo(dtype).eps))


def test_struct_array():
    n = 2
    m = 3

    np_nested1 = np.dtype([("a", np.int64, (n, n)), ("b", np.float64, (m,))], align=True)
    np_nested2 = np.dtype([("c", np_nested1, (n,)), ("d", np_nested1, (3,))], align=True)
    np_nested3 = np.dtype([("c", np_nested2, (2,)), ("d", np_nested1, (3,))], align=True)

    @nm.jit
    def mod_struct(a) -> None:
        a[1]["c"][1]["d"][2]["b"][1] = -4.0

    a = np.zeros(2, dtype=np_nested3)

    mod_struct(a)

    b = np.zeros(2, dtype=np_nested3)
    b[1]["c"][1]["d"][2]["b"][1] = -4.0

    np.testing.assert_equal(a, b)


def test_struct():
    n = 2
    m = 3

    np_nested1 = np.dtype([("a", np.int64, (n, n)), ("b", np.float64, (m,))], align=True)
    np_nested2 = np.dtype([("c", np_nested1, (n,)), ("d", np_nested1, (3,))], align=True)
    np_nested3 = np.dtype([("c", np_nested2, (2,)), ("d", np_nested1, (3,))], align=True)

    @nm.jit
    def mod_struct(a) -> None:
        a["c"][1]["d"][2]["b"][1] = -4.0

    a = np.zeros(2, dtype=np_nested3)

    mod_struct(a[1])

    b = np.zeros(2, dtype=np_nested3)
    b[1]["c"][1]["d"][2]["b"][1] = -4.0

    np.testing.assert_equal(a, b)



import numpy as np
import numeta as nm


def test_declare_global_constant():

    global_constant_var = nm.declare_global_constant(
        (2, 1), np.float64, value=np.array([2.0, -1.0]), name="global_constant_var"
    )

    @nm.jit
    def get_global_constant(var):
        var[0] = global_constant_var[0, 0]
        var[1] = global_constant_var[1, 0]

    a = np.empty(2, dtype=np.float64)
    get_global_constant(a)

    np.testing.assert_allclose(a, np.array([2.0, -1.0]))


def test_declare_global_constant_nested():

    global_constant_var = nm.declare_global_constant(
        (2, 1), np.float64, value=np.array([2.0, -1.0]), name="global_constant_var"
    )

    @nm.jit
    def get_global_constant_nested(var):
        var[0] = global_constant_var[0, 0]
        var[1] = global_constant_var[1, 0]

    @nm.jit
    def get_global_constant(var):
        get_global_constant_nested(var)

    a = np.empty(2, dtype=np.float64)
    get_global_constant(a)

    np.testing.assert_allclose(a, np.array([2.0, -1.0]))



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_mul(dtype):
    n = 100

    @nm.jit
    def mul(a, b, c):
        for i in nm.range(a.shape[0]):
            for k in nm.range(b.shape[0]):
                c[i, :] += a[i, k] * b[k, :]

    a = np.random.rand(n, n).astype(dtype)
    b = np.random.rand(n, n).astype(dtype)
    c = np.zeros((n, n), dtype=dtype)

    mul(a, b, c)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(c, np.dot(a, b), atol=0)
    else:
        np.testing.assert_allclose(c, np.dot(a, b), rtol=10e2 * np.finfo(dtype).eps)



import numpy as np

from numeta.signature import (
    convert_signature_to_argument_specs,
    get_signature_and_runtime_args,
    parse_function_parameters,
)
from numeta.types_hint import comptime


def test_parse_parameters_with_varargs_and_kwargs():
    def sample(a, *values, b=1, **options):
        return a, values, b, options

    (
        params,
        fixed_param_indices,
        n_positional_or_default_args,
        catch_var_positional_name,
    ) = parse_function_parameters(sample)

    assert catch_var_positional_name == "values"
    assert n_positional_or_default_args == 2
    assert [params[idx].name for idx in fixed_param_indices] == ["a", "b"]


def test_signature_parsing_with_comptime_and_keyword_only():
    def sample(a: comptime, b, *, c=2):
        return a, b, c

    (
        params,
        fixed_param_indices,
        n_positional_or_default_args,
        catch_var_positional_name,
    ) = parse_function_parameters(sample)

    array = np.arange(4, dtype=np.float32)
    to_execute, signature, runtime_args = get_signature_and_runtime_args(
        (3, array),
        {"c": 5},
        params=params,
        fixed_param_indices=fixed_param_indices,
        n_positional_or_default_args=n_positional_or_default_args,
        catch_var_positional_name=catch_var_positional_name,
    )

    assert to_execute is True
    assert signature[0] == 3
    assert signature[1][0] == "b"
    assert signature[1][1] == array.dtype
    assert signature[2][0] == "c"
    assert runtime_args == [array, 5]

    argument_specs = convert_signature_to_argument_specs(
        signature,
        params=params,
        fixed_param_indices=fixed_param_indices,
        n_positional_or_default_args=n_positional_or_default_args,
    )
    assert argument_specs[0].is_comptime is True
    assert argument_specs[1].name == "b"
    assert argument_specs[2].name == "c"
    assert argument_specs[2].is_keyword is True


def test_signature_parsing_with_varargs():
    def sample(a, *vals):
        return a, vals

    (
        params,
        fixed_param_indices,
        n_positional_or_default_args,
        catch_var_positional_name,
    ) = parse_function_parameters(sample)

    vector = np.arange(3, dtype=np.int32)
    to_execute, signature, runtime_args = get_signature_and_runtime_args(
        (1, vector),
        {},
        params=params,
        fixed_param_indices=fixed_param_indices,
        n_positional_or_default_args=n_positional_or_default_args,
        catch_var_positional_name=catch_var_positional_name,
    )

    assert to_execute is True
    assert signature[0] == ("a", int)
    assert signature[1][0] == ("vals", 0)
    assert runtime_args == [1, vector]

    argument_specs = convert_signature_to_argument_specs(
        signature,
        params=params,
        fixed_param_indices=fixed_param_indices,
        n_positional_or_default_args=n_positional_or_default_args,
    )
    assert argument_specs[0].name == "a"
    assert argument_specs[1].name == "vals_0"
    assert argument_specs[1].is_keyword is False



import numpy as np
from numeta.builder_helper import BuilderHelper
from numeta.datatype import DataType, FortranType, get_datatype


def scalar(dtype: DataType | FortranType | np.generic, value=None, name=None):
    if isinstance(dtype, FortranType):
        ftype = dtype
    else:
        ftype = get_datatype(dtype).get_fortran()

    var = BuilderHelper.generate_local_variables("fc_s", ftype=ftype, name=name)
    if value is not None:
        var[:] = value
    return var



import inspect
from dataclasses import dataclass
from typing import Any

import numpy as np

from .array_shape import ArrayShape, SCALAR, UNKNOWN
from .datatype import DataType, ArrayType, get_datatype
from .settings import settings
from .syntax import Variable
from .syntax.expressions import ExpressionNode, GetAttr, GetItem
from .types_hint import comptime


@dataclass(frozen=True)
class ArgumentSpec:
    """
    This class is used to store the details of the arguments of the function.
    The ones that are compile-time are stored in the is_comptime attribute.
    """

    name: str
    is_comptime: bool = False
    comptime_value: Any = None
    datatype: DataType | None = None
    shape: ArrayShape | None = None
    rank: int = 0  # rank of the array, 0 for scalar
    intent: str = "inout"  # can be "in" or "inout"
    to_pass_by_value: bool = False
    is_keyword: bool = False


@dataclass(frozen=True)
class ParameterInfo:
    """Store metadata about a Python function parameter."""

    name: str
    kind: inspect._ParameterKind
    default: Any = inspect._empty
    is_comptime: bool = False


def parse_function_parameters(func):
    py_signature = inspect.signature(func)

    params = []
    catch_var_positional_name = "args"

    for name, parameter in py_signature.parameters.items():
        is_comp = func.__annotations__.get(name) is comptime
        params.append(ParameterInfo(name, parameter.kind, parameter.default, is_comp))

        if parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            catch_var_positional_name = name

    fixed_param_indices = [
        i
        for i, p in enumerate(params)
        if p.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
    ]
    n_positional_or_default_args = len(fixed_param_indices)

    return params, fixed_param_indices, n_positional_or_default_args, catch_var_positional_name


def get_signature_and_runtime_args(
    args,
    kwargs,
    *,
    params,
    fixed_param_indices,
    n_positional_or_default_args,
    catch_var_positional_name,
):
    """
    This method quickly extracts the signature and runtime arguments from the provided args.
    If the runtime arguments are not all numpy arrays or numeric types, the call is not to be
    executed.
    It returns a tuple of:
    - to_execute: a boolean indicating if the function can be executed with the provided args
    - signature: a tuple of the signature of the function
    - runtime_args: a list of runtime arguments to be passed to run the function
    A signature is a tuple of tuples, where each inner tuple represents an argument.
    - (name, dtype,) is scalar types passed by value
    - (name, dtype, 0) is for scalar types passed by reference
    - (name, dtype, rank) for numpy arrays
    - (name, dtype, rank, has_fortran_order) to set the Fortran order
    - (name, dtype, rank, has_fortran_order, intent) intent can be "in" or "inout"
    - (name, dtype, rank, has_fortran_order, intent, shape) if the shape is know at comptime
    **name** is a tuple is the argument comes from a variable positional argument (*args)
    """

    to_execute = True

    def get_signature_from_arg(arg, name):
        nonlocal to_execute

        if isinstance(arg, np.ndarray):
            arg_signature = (name, arg.dtype, len(arg.shape), np.isfortran(arg))
        elif isinstance(arg, (int, float, complex)):
            arg_signature = (
                name,
                type(arg),
            )
        elif isinstance(arg, np.generic):
            # it is a numpy scalar like np.int32(1) or np.float64(1.0) or a struct
            # A struct is mutable
            if arg.dtype.names is not None:
                arg_signature = (name, arg.dtype, 0)
            else:
                arg_signature = (
                    name,
                    arg.dtype,
                )
        elif isinstance(arg, ArrayType):
            to_execute = False
            if arg.shape is UNKNOWN or (
                not settings.add_shape_descriptors and arg.shape.has_comptime_undefined_dims()
            ):
                # it is a pointer
                arg_signature = (name, arg.dtype.get_numpy(), None, arg.shape.fortran_order)
            elif arg.shape.has_comptime_undefined_dims():
                arg_signature = (
                    name,
                    arg.dtype.get_numpy(),
                    arg.shape.rank,
                    arg.shape.fortran_order,
                )
            else:
                arg_signature = (
                    name,
                    arg.dtype.get_numpy(),
                    arg.shape.rank,
                    arg.shape.fortran_order,
                    "inout",
                    arg.shape.dims,
                )
        elif isinstance(arg, type) and issubclass(arg, DataType):
            to_execute = False
            arg_signature = (
                name,
                arg.get_numpy(),
            )
        elif isinstance(arg, ExpressionNode):
            to_execute = False
            ftype = arg._ftype
            dtype = DataType.from_ftype(ftype)
            # Let's stay safe, let's assume is an expression so intent is in
            intent = "in"
            # These are the cases where we can assume it is an inout argument
            # becase the intent can be only "in" or "inout"
            if isinstance(arg, Variable) and arg.intent != "in":
                intent = "inout"
            if isinstance(arg, GetAttr) and arg.variable.intent != "in":
                intent = "inout"
            if isinstance(arg, GetItem) and arg.variable.intent != "in":
                intent = "inout"

            if arg._shape is SCALAR:
                if intent == "inout":
                    arg_signature = (name, dtype.get_numpy(), 0, False, intent)
                else:
                    arg_signature = (
                        name,
                        dtype.get_numpy(),
                    )
            elif arg._shape is UNKNOWN or (
                not settings.add_shape_descriptors and arg._shape.has_comptime_undefined_dims()
            ):
                arg_signature = (name, dtype.get_numpy(), None, False, intent)
            elif arg._shape.has_comptime_undefined_dims():
                arg_signature = (
                    name,
                    dtype.get_numpy(),
                    arg._shape.rank,
                    arg._shape.fortran_order,
                    intent,
                )
            else:
                if not settings.ignore_fixed_shape_in_nested_calls:
                    arg_signature = (
                        name,
                        dtype.get_numpy(),
                        arg._shape.rank,
                        arg._shape.fortran_order,
                        intent,
                        arg._shape.dims,
                    )
                else:
                    arg_signature = (name, dtype.get_numpy(), None, False, intent)
        else:
            raise ValueError(f"Argument {name} of type {type(arg)} is not supported")

        return arg_signature

    runtime_args = []
    signature = [None] * n_positional_or_default_args

    unused_kwargs = kwargs
    pos_idx = 0

    for fi, param_idx in enumerate(fixed_param_indices):
        param = params[param_idx]

        if param.kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ):
            if pos_idx < len(args):
                arg = args[pos_idx]
                pos_idx += 1
            elif param.name in unused_kwargs:
                arg = unused_kwargs.pop(param.name)
            elif param.default is not inspect._empty:
                arg = param.default
            else:
                raise ValueError(f"Missing required argument: {param.name}")
        else:  # KEYWORD_ONLY
            if param.name in unused_kwargs:
                arg = unused_kwargs.pop(param.name)
            elif param.default is not inspect._empty:
                arg = param.default
            else:
                raise ValueError(f"Missing required argument: {param.name}")

        if param.is_comptime:
            signature[fi] = arg
        else:
            signature[fi] = get_signature_from_arg(arg, param.name)
            runtime_args.append(arg)

    # catch the *args variable positional arguments
    if pos_idx < len(args):
        for j, arg in enumerate(args[pos_idx:]):
            name = (catch_var_positional_name, j)
            signature.append(get_signature_from_arg(arg, name))
            runtime_args.append(arg)

    # catch the **kwargs variable keyword arguments
    unused_kwargs_keys = (
        unused_kwargs.keys() if not settings.reorder_kwargs else sorted(unused_kwargs.keys())
    )
    for name in unused_kwargs_keys:
        arg = unused_kwargs[name]
        signature.append(get_signature_from_arg(arg, name))
        runtime_args.append(arg)

    return to_execute, tuple(signature), runtime_args


def convert_signature_to_argument_specs(
    signature,
    *,
    params,
    fixed_param_indices,
    n_positional_or_default_args,
):
    """
    Converts a signature tuple into a list of ArgumentSpec objects.
    A signature is a tuple of tuples, where each inner tuple represents an argument.
    """

    def convert_arg_to_argument_spec(arg, is_keyword, name=None):
        name = arg[0] if name is None else name

        dtype = get_datatype(arg[1])
        if len(arg) == 2:
            # it is a numeric type or a string
            # So the intent will be always "in"
            # but complex numbers cannot be passed by value because of C
            ap = ArgumentSpec(
                name,
                datatype=dtype,
                shape=SCALAR,
                to_pass_by_value=dtype.can_be_value(),
                intent="in",
                is_keyword=is_keyword,
            )
        else:
            # for numpy arrays arg[1] is the rank, for the other types it is the shape
            rank = arg[2]

            fortran_order = False
            if len(arg) == 4:
                fortran_order = arg[3]

            intent = "inout"
            if len(arg) == 5:
                intent = arg[4]

            if rank is None:
                shape = UNKNOWN
            elif rank == 0:
                shape = SCALAR
            else:
                shape = ArrayShape([None] * rank, fortran_order=fortran_order)

            if len(arg) == 6:
                # it means that the shape is known at comptime
                shape = ArrayShape(arg[5], fortran_order=fortran_order)

            ap = ArgumentSpec(
                name,
                datatype=dtype,
                rank=rank,
                shape=shape,
                intent=intent,
                is_keyword=is_keyword,
            )

        return ap

    signature_spec = []
    for i, arg in enumerate(signature):

        if i < n_positional_or_default_args:
            param = params[fixed_param_indices[i]]
            if param.is_comptime:
                ap = ArgumentSpec(
                    param.name,
                    is_comptime=True,
                    comptime_value=arg,
                    is_keyword=param.kind == inspect.Parameter.KEYWORD_ONLY,
                )
            else:
                is_keyword = param.kind == inspect.Parameter.KEYWORD_ONLY
                ap = convert_arg_to_argument_spec(arg, is_keyword)
        elif isinstance(arg[0], tuple):
            # it is a positional argument called with *args
            is_keyword = False
            name = f"{arg[0][0]}_{arg[0][1]}"
            ap = convert_arg_to_argument_spec(arg, is_keyword, name=name)
        else:
            is_keyword = True
            ap = convert_arg_to_argument_spec(arg, is_keyword)

        signature_spec.append(ap)

    return signature_spec



import numeta as nm


def test_datatype_getitem():
    arr = nm.int32[2, 3]
    assert isinstance(arr, nm.datatype.ArrayType)
    assert arr.dtype is nm.int32
    assert arr.shape.dims == (2, 3)

    colon_arr = nm.float64[:]
    assert colon_arr.shape.dims == (None,)



from string import Template
import subprocess as sp
import sysconfig
import numpy as np

from .settings import settings


class CAPIInterface:
    def __init__(
        self,
        name,
        module_name,
        args_details,
        return_specs,
        directory,
        compile_flags,
        do_checks=True,
        compiler="gcc",
    ):
        self.name = name
        self.module_name = module_name
        self.args_details = args_details
        self.return_specs = return_specs
        self.directory = directory
        if isinstance(compile_flags, str):
            self.compile_flags = compile_flags.split()
        else:
            self.compile_flags = compile_flags
        self.do_checks = do_checks
        self.compiler = compiler

    def generate(self):
        capi_interface = self.construct_module()
        capi_interface_src = self.directory / f"{self.module_name}.c"
        capi_interface_src.write_text(capi_interface)
        capi_obj_path = self.compile(capi_interface_src)
        return capi_obj_path

    def construct_module(self):
        template = """
#include <Python.h>
#include <numpy/arrayobject.h>

void numpy_allocate(void **ptr, size_t *nbytes) {
    *ptr = PyDataMem_NEW(*nbytes);
}

void numpy_deallocate(void **ptr) {
    PyDataMem_FREE(*ptr);
}

${struct_definitions}

${procedure_definitions}

// Method definition table
static PyMethodDef Methods[] = {
    ${module_methods}
    {NULL, NULL, 0, NULL}  // Sentinel
};

// Module definition
static struct PyModuleDef ${module_name} = {
    PyModuleDef_HEAD_INIT,
    "${module_name}",  // Module name
    NULL,  // Optional module documentation
    -1,
    Methods
};

// Module initialization function
PyMODINIT_FUNC PyInit_${module_name}(void) {
    import_array();  // Initialize NumPy API
    return PyModule_Create(&${module_name});
}
"""

        substitutions = {}
        substitutions["module_name"] = f"{self.module_name}"

        struct_definitions = {}
        for arg in self.args_details:
            if not arg.is_comptime and arg.datatype.is_struct():
                # bfs to declare all nested structs
                def bfs(struct):
                    for _, nested_dtype, _ in struct.members:
                        if nested_dtype.is_struct() and nested_dtype.name not in struct_definitions:
                            bfs(nested_dtype)
                    struct_definitions[struct.name] = struct.c_declaration()

                bfs(arg.datatype)
        substitutions["struct_definitions"] = "".join(struct_definitions.values())

        module_methods = [
            f'{{"{self.name}", (PyCFunction)(void(*)(void))fc_{self.name}, METH_FASTCALL, "Wrapper generated by fortranic"}},'
        ]
        substitutions["module_methods"] = "".join(module_methods)

        procedure_definitions = [self.construct_procedure()]
        substitutions["procedure_definitions"] = "\n".join(procedure_definitions)

        module_template = Template(template)
        module_template = module_template.substitute(substitutions)

        return module_template

    def construct_procedure(self):
        template = """
void ${fortran_name}(${fortran_args});

static PyObject* ${procedure_name}(PyObject *self, PyObject *const *args, Py_ssize_t nargs) {

    if (nargs != ${n_args}) {
        PyErr_SetString(PyExc_TypeError, "${fortran_name} takes exactly ${n_args} arguments");
        return NULL;
    }

    ${initializations}

    if (PyErr_Occurred()) {
        PyErr_SetString(PyExc_TypeError, "Error occurred while parsing arguments.");
        return NULL;
    }

    ${checks}

    ${return_args_declarations}

    ${fortran_name}(${call_args});

    ${return_args_conversion_to_numpy}

    ${return}
}
"""
        args = [a for a in self.args_details if not a.is_comptime]

        substitutions = {}
        substitutions["fortran_name"] = self.name
        substitutions["procedure_name"] = f"fc_{self.name}"
        substitutions["n_args"] = len(args)

        fortran_args = [self.get_fortran_args(var) for var in args]
        substitutions["initializations"] = "\n    ".join(
            [self.get_initialization(i, var) for i, var in enumerate(args)]
        )

        substitutions["checks"] = ""
        if self.do_checks:
            substitutions["checks"] = "\n    ".join([self.get_check(var) for var in args])

        call_args = [self.get_call_args(var) for var in args]

        substitutions["return_args_declarations"] = "\n"
        substitutions["return_args_conversion_to_numpy"] = ""
        return_vars = []
        for i, (dtype, rank) in enumerate(self.return_specs):
            if rank != 0:
                substitutions[
                    "return_args_declarations"
                ] += f"\n    npy_intp out_shape_{i}[{rank}];"
                call_args.append(f"out_shape_{i}")
                fortran_args.append(f"npy_intp* out_shape_{i}")

                substitutions["return_args_declarations"] += f"\n    void* out_ptr_{i} = NULL;"
                call_args.append(f"&out_ptr_{i}")
                fortran_args.append(f"void** out_ptr_{i}")
                substitutions[
                    "return_args_conversion_to_numpy"
                ] += f"\n    PyObject* ret_out_{i} = PyArray_SimpleNewFromData({rank}, out_shape_{i}, {dtype.get_cnumpy().upper()}, out_ptr_{i});"
                substitutions[
                    "return_args_conversion_to_numpy"
                ] += f"\n    PyArray_ENABLEFLAGS((PyArrayObject*)ret_out_{i}, NPY_ARRAY_OWNDATA);"
                return_vars.append(f"ret_out_{i}")
            else:
                substitutions[
                    "return_args_declarations"
                ] += f"\n    {dtype.get_cnumpy()} out_val_{i};"
                call_args.append(f"&out_val_{i}")
                fortran_args.append(f"{dtype.get_cnumpy()}* out_val_{i}")
                substitutions[
                    "return_args_conversion_to_numpy"
                ] += f"\n    PyObject* ret_out_{i} = PyArray_Scalar(&out_val_{i}, PyArray_DescrFromType({dtype.get_cnumpy().upper()}), NULL);"
                return_vars.append(f"ret_out_{i}")

        substitutions["fortran_args"] = ", ".join(fortran_args)
        substitutions["call_args"] = ", ".join(call_args)

        if len(return_vars) == 0:
            substitutions["return"] = "Py_RETURN_NONE;"
        elif len(return_vars) == 1:
            substitutions["return"] = f"return {return_vars[0]};"
        else:
            return_str = f"PyObject* tuple = PyTuple_New({len(return_vars)});"
            for i, var in enumerate(return_vars):
                return_str += f"\n    Py_INCREF({var});"
                return_str += f"\n    PyTuple_SetItem(tuple, {i}, {var});"
            return_str += "\n    return tuple;"
            substitutions["return"] = return_str

        return Template(template).substitute(substitutions)

    def get_fortran_args(self, variable):
        """
        Returns the Fortran argument declaration for a given variable.
        Example for this function:

            void matmul(int32_t* a, int32_t* b, int32_t* c);

        this function returns the string 'int32_t* a' for the first argument.
        """

        if variable.to_pass_by_value:
            return f"{variable.datatype.get_cnumpy()} {variable.name}"
        elif settings.add_shape_descriptors and variable.shape.has_comptime_undefined_dims():
            return (
                f"npy_intp* {variable.name}_dims, {variable.datatype.get_cnumpy()}* {variable.name}"
            )
        else:
            return f"{variable.datatype.get_cnumpy()}* {variable.name}"

    def get_call_args(self, variable):
        """
        Returns the call argument declaration for a given variable.
        Example for this function:

            ciao((int32_t)a);

        this function returns the string '(int32_t)a' for the first argument.
        """
        if variable.to_pass_by_value:
            return f"({variable.datatype.get_cnumpy()}){variable.name}"
        elif variable.rank == 0 and variable.datatype.is_struct():
            return f"({variable.datatype.get_cnumpy()}*){variable.name}"
        elif settings.add_shape_descriptors and variable.shape.has_comptime_undefined_dims():
            return f"PyArray_DIMS({variable.name}), ({variable.datatype.get_cnumpy()}*)PyArray_DATA({variable.name})"
        else:
            return f"({variable.datatype.get_cnumpy()}*)PyArray_DATA({variable.name})"

    def get_initialization(self, i, variable):
        """
        Returns the declaration and initialization of the variable in the C API interface.
        For instance:

            int32_t a = (int32_t)PyLong_AsLongLong(args[0]);

        Where args[idx] is the corresponding argument in the Python function.
        """

        if variable.to_pass_by_value:
            if variable.datatype.get_numpy() in (np.complex64, np.complex128):
                result = "\n"
                result += "#if NPY_ABI_VERSION < 0x02000000\n"
                result += f"    {variable.datatype.get_cnumpy()} {variable.name};\n"
                result += f"    {variable.name}.real = PyComplex_RealAsDouble(args[{i}]);\n"
                result += f"    {variable.name}.imag = PyComplex_RealAsDouble(args[{i}]);\n"
                result += "#else\n"
                cast = variable.datatype.get_capi_cast(f"args[{i}]")
                result += f"    {variable.datatype.get_cnumpy()} {variable.name} = ({variable.datatype.get_cnumpy()}){cast};\n"
                result += "#endif"
                return result
            cast = variable.datatype.get_capi_cast(f"args[{i}]")
            return f"{variable.datatype.get_cnumpy()} {variable.name} = ({variable.datatype.get_cnumpy()}){cast};"
        elif variable.rank == 0 and variable.datatype.is_struct():
            result = f"{variable.datatype.get_cnumpy()}* {variable.name} = NULL;\n"
            result += f"     PyArray_ScalarAsCtype(args[{i}], &{variable.name});"
            return result
        return f"PyArrayObject *{variable.name} = (PyArrayObject*)args[{i}];"

    def get_check(self, variable):
        if variable.rank != 0:
            check_array = f"""
    if (!PyArray_Check({variable.name})) {{
        PyErr_SetString(PyExc_TypeError, "Argument '{variable.name}' must be a NumPy array");
        return NULL;
    }}"""
            if variable.shape.fortran_order:
                check_farray = f"""
    if (!PyArray_ISFARRAY({variable.name})) {{
        PyErr_SetString(PyExc_ValueError, "Input array '{variable.name}' is not Fortran contiguous or aligned.");
        return NULL; 
    }}"""
            else:
                check_farray = f"""
    if (!PyArray_ISCARRAY({variable.name})) {{
        PyErr_SetString(PyExc_ValueError, "Input array '{variable.name}' is not C contiguous or aligned.");
        return NULL; 
    }}"""
            if variable.datatype.is_struct():
                check_align = f"""
    if (!PyDataType_FLAGCHK((PyArray_Descr*)PyArray_DESCR({variable.name}), NPY_ALIGNED_STRUCT)) {{
        PyErr_SetString(PyExc_ValueError, "Input struct {variable.name} dtype is not aligned, use align=True ");
        return NULL;
    }}"""
                return check_array + check_farray + check_align
            else:
                check_type = f"""
    if (PyArray_TYPE({variable.name}) != {variable.datatype.get_cnumpy().upper()}) {{
        PyErr_SetString(PyExc_ValueError, "Input array '{variable.name}' does not have the required type ({variable.datatype.get_numpy()}) "); 
        return NULL;
    }}"""
                return check_array + check_farray + check_type
        else:
            return ""

    def compile(
        self,
        capi_interface_src,
    ):
        """
        Compiles the C API interface.

        Parameters:
            capi_interface_src (str): Source code of the C API interface.

        Returns:
            Path: Path to the compiled C API interface object file.
        """

        include_dirs = [sysconfig.get_paths()["include"], np.get_include()]

        capi_obj_path = self.directory / f"{self.name}_capi.o"

        command = [self.compiler]
        command.extend(self.compile_flags)
        command.extend(["-fPIC", "-c", "-o", str(capi_obj_path), str(capi_interface_src)])
        command.extend([f"-I{inc_dir}" for inc_dir in include_dirs])
        command.extend(["-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION"])

        try:
            sp.run(command, check=True, cwd=self.directory)
        except sp.CalledProcessError as e:
            print(f"Compilation failed with error: {e}")
            raise

        return capi_obj_path



import pytest
import numpy as np

import numeta as nm
from numeta.array_shape import ArrayShape, SCALAR
from numeta.syntax import Variable, Assignment, LiteralNode, DerivedType
from numeta.syntax.expressions import GetAttr, Function
from numeta.syntax import Do, DoWhile, If, ElseIf, Else
from numeta.syntax.statements.tools import print_block
from numeta.syntax.expressions import (
    Abs,
    Neg,
    Not,
    Allocated,
    Shape,
    All,
    Real,
    Imag,
    Complex,
    Conjugate,
    Transpose,
    Exp,
    Sqrt,
    Floor,
    Sin,
    Cos,
    Tan,
    Sinh,
    Cosh,
    Tanh,
    ASin,
    ACos,
    ATan,
    ATan2,
    Dotproduct,
    Rank,
    Size,
    Max,
    Maxval,
    Min,
    Minval,
    Iand,
    Ior,
    Xor,
    Ishft,
    Ibset,
    Ibclr,
    Popcnt,
    Trailz,
    Sum,
    Matmul,
    ArrayConstructor,
)
from numeta.syntax.statements import VariableDeclaration, Call
from numeta.syntax import Subroutine, Module, Scope
from numeta.syntax.settings import settings as syntax_settings
from numeta.settings import settings


def render(expr):
    """Return a string representation of an expression."""
    return print_block(expr.get_code_blocks())


def test_simple_assignment_syntax():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    stmt = Assignment(x, y, add_to_scope=False)
    assert stmt.print_lines() == ["x=y\n"]


def test_literal_node():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    lit = LiteralNode(5)
    assert render(lit) == "5_c_int64_t\n"


def test_binary_operation_node():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    expr = x + y
    assert render(expr) == "(x+y)\n"


def test_getattr_node():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    expr = GetAttr(x, "tag")
    assert render(expr) == "x%tag\n"


def test_getitem_node():
    arr = Variable("a", syntax_settings.DEFAULT_REAL, shape=(10, 10))
    expr = arr[1, 2]
    assert render(expr) == "a(1, 2)\n"


def test_unary_neg_node():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    expr = -x
    assert render(expr) == "-(x)\n"


def test_eq_ne_nodes():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    assert render(x == y) == "(x.eq.y)\n"
    assert render(x != y) == "(x.ne.y)\n"


def test_re_im_nodes():
    z = Variable("z", syntax_settings.DEFAULT_COMPLEX)
    assert render(z.real) == "z%re\n"
    assert render(z.imag) == "z%im\n"


def test_array_constructor():
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    arr = Variable("arr", syntax_settings.DEFAULT_INTEGER, shape=(10, 10))
    expr = ArrayConstructor(arr[1, 1], 5, i).get_code_blocks()
    expected = ["[", "arr", "(", "1", ",", " ", "1", ")", ", ", "5_c_int64_t", ", ", "i", "]"]
    assert expr == expected


def test_complex_function_default():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    settings.set_default_from_datatype(nm.complex128, iso_c=True)

    a = Variable("a", syntax_settings.DEFAULT_REAL)
    b = Variable("b", syntax_settings.DEFAULT_REAL)
    expr = Complex(a, b)
    assert render(expr) == "cmplx(a, b, c_double_complex)\n"


def test_complex_function():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    settings.set_default_from_datatype(nm.complex128, iso_c=True)
    settings.set_default_from_datatype(nm.int64, iso_c=True)

    a = Variable("a", syntax_settings.DEFAULT_REAL)
    b = Variable("b", syntax_settings.DEFAULT_REAL)
    expr = Complex(a, b, kind=8)
    assert render(expr) == "cmplx(a, b, 8_c_int64_t)\n"


@pytest.mark.parametrize(
    "func,nargs,token",
    [
        (Abs, 1, "abs"),
        (Neg, 1, "-"),
        (Not, 1, ".not."),
        (Allocated, 1, "allocated"),
        # TODO(Shape, 1, "shape"),
        (All, 1, "all"),
        (Real, 1, "real"),
        (Imag, 1, "aimag"),
        (Conjugate, 1, "conjg"),
        (Transpose, 1, "transpose"),
        (Exp, 1, "exp"),
        (Sqrt, 1, "sqrt"),
        (Floor, 1, "floor"),
        (Sin, 1, "sin"),
        (Cos, 1, "cos"),
        (Tan, 1, "tan"),
        (Sinh, 1, "sinh"),
        (Cosh, 1, "cosh"),
        (Tanh, 1, "tanh"),
        (ASin, 1, "asin"),
        (ACos, 1, "acos"),
        (ATan, 1, "atan"),
        (Rank, 1, "rank"),
        (Maxval, 1, "maxval"),
        (Minval, 1, "minval"),
        (Popcnt, 1, "popcnt"),
        (Trailz, 1, "trailz"),
        (Sum, 1, "sum"),
        (ATan2, 2, "atan2"),
        (Dotproduct, 2, "dot_product"),
        (Size, 2, "size"),
        (Max, 2, "max"),
        (Min, 2, "min"),
        (Iand, 2, "iand"),
        (Ior, 2, "ior"),
        (Xor, 2, "xor"),
        (Ishft, 2, "ishft"),
        (Ibset, 2, "ibset"),
        (Ibclr, 2, "ibclr"),
        (Matmul, 2, "matmul"),
    ],
)
def test_intrinsic_functions(func, nargs, token):
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    args = [x] if nargs == 1 else [x, y]
    if func is Size:
        args[1] = 1
    expr = func(*args)
    expected_args = ["x"] if nargs == 1 else ["x", "y"]
    if func is Size:
        expected_args[1] = "1_c_int64_t"
    expected = f"{token}({', '.join(expected_args)})\n"
    assert render(expr) == expected


def test_variable_declaration_scalar():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    dec = VariableDeclaration(x)
    assert dec.print_lines() == ["integer(c_int64_t) :: x\n"]


def test_variable_declaration_array():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    a = Variable("a", syntax_settings.DEFAULT_REAL, shape=(5,))
    dec = VariableDeclaration(a)
    assert dec.print_lines() == ["real(c_double), dimension(0:4) :: a\n"]


def test_variable_declaration_pointer():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    p = Variable("p", syntax_settings.DEFAULT_REAL, shape=(10, 10), pointer=True)
    dec = VariableDeclaration(p)
    assert dec.print_lines() == ["real(c_double), pointer, dimension(:,:) :: p\n"]


def test_variable_declaration_allocatable():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    arr = Variable("arr", syntax_settings.DEFAULT_REAL, shape=(3, 3), allocatable=True)
    dec = VariableDeclaration(arr)
    assert dec.print_lines() == ["real(c_double), allocatable, dimension(:,:) :: arr\n"]


def test_variable_declaration_intent():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    v = Variable("v", syntax_settings.DEFAULT_REAL, intent="in")
    dec = VariableDeclaration(v)
    assert dec.print_lines() == ["real(c_double), intent(in), value :: v\n"]


def test_variable_declaration_bind_c():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    v = Variable("v", syntax_settings.DEFAULT_REAL, bind_c=True)
    dec = VariableDeclaration(v)
    assert dec.print_lines() == ["real(c_double), bind(C, name='v') :: v\n"]


def test_variable_declaration_assign_scalar():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    v = Variable("v", syntax_settings.DEFAULT_REAL, assign=5.0)
    dec = VariableDeclaration(v)
    assert dec.print_lines() == ["real(c_double) :: v; data v / 5.0_c_double /\n"]


def test_variable_declaration_assign_array():
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    v = Variable("v", syntax_settings.DEFAULT_REAL, shape=(2, 1), assign=np.array([3.0, 5.0]))
    dec = VariableDeclaration(v)
    assert dec.print_lines() == [
        "real(c_double), dimension(0:1, 0:0) :: v; data v / 3.0_c_double, 5.0_c_double /\n"
    ]


def test_subroutine_print_lines():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER, intent="in")
    y = Variable("y", syntax_settings.DEFAULT_INTEGER, intent="out")
    sub = Subroutine("mysub")
    sub.add_variable(x, y)
    with sub.scope:
        Assignment(y, x)
    expected = [
        "subroutine mysub(x, y) bind(C, name='mysub')\n",
        "    use iso_c_binding, only: c_int64_t\n",
        "    implicit none\n",
        "    integer(c_int64_t), intent(in), value :: x\n",
        "    integer(c_int64_t), intent(out) :: y\n",
        "    y=x\n",
        "end subroutine mysub\n",
    ]
    assert sub.print_lines() == expected


def test_module_print_code():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER, intent="in")
    mod = Module("mymod")
    sub = Subroutine("mysub", parent=mod)
    sub.add_variable(x)
    expected = [
        "module mymod\n",
        "    implicit none\n",
        "    contains\n",
        "    subroutine mysub(x) bind(C, name='mysub')\n",
        "        use iso_c_binding, only: c_int64_t\n",
        "        implicit none\n",
        "        integer(c_int64_t), intent(in), value :: x\n",
        "    end subroutine mysub\n",
        "end module mymod\n",
    ]
    assert mod.print_lines() == expected


def test_derived_type_declaration():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    settings.set_default_from_datatype(nm.float64, iso_c=True)
    dt = DerivedType(
        "point",
        [
            ("x", syntax_settings.DEFAULT_INTEGER, SCALAR),
            ("y", syntax_settings.DEFAULT_INTEGER, SCALAR),
            ("arr", syntax_settings.DEFAULT_REAL, ArrayShape((5,))),
        ],
    )
    expected = [
        "type, bind(C) :: point\n",
        "    integer(c_int64_t) :: x\n",
        "    integer(c_int64_t) :: y\n",
        "    real(c_double), dimension(0:4) :: arr\n",
        "end type point\n",
    ]
    assert dt.get_declaration().print_lines() == expected


def test_do_statement():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)

    do = Do(i, 0, 3, add_to_scope=False)
    with do:
        Assignment(x, i + 1)

    expected = ["do i = 0_c_int64_t, 3_c_int64_t\n", "    x=(i+1_c_int64_t)\n", "end do\n"]

    for l1, l2 in zip(do.print_lines(), expected):
        assert l1 == l2


def test_if_statement():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)

    wrapper = Do(i, 0, 3, add_to_scope=False)
    with wrapper:
        with If(i < 5):
            Assignment(x, i + 1)
        with ElseIf(i < 10):
            Assignment(x, i + 2)
        with Else():
            Assignment(x, 0)

    expected = [
        "do i = 0_c_int64_t, 3_c_int64_t\n",
        "    if((i.lt.5_c_int64_t))then\n",
        "        x=(i+1_c_int64_t)\n",
        "    elseif((i.lt.10_c_int64_t))then\n",
        "        x=(i+2_c_int64_t)\n",
        "    else\n",
        "        x=0_c_int64_t\n",
        "    end if\n",
        "end do\n",
    ]

    for l1, l2 in zip(wrapper.print_lines(), expected):
        assert l1 == l2


def test_do_while_statement():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)

    do = DoWhile(i < 5, add_to_scope=False)
    with do:
        Assignment(x, i + 1)

    expected = ["do while ((i.lt.5_c_int64_t))\n", "    x=(i+1_c_int64_t)\n", "end do\n"]

    for l1, l2 in zip(do.print_lines(), expected):
        assert l1 == l2


def test_update_variables_simple_assignment():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    stmt = Assignment(x, y, add_to_scope=False)

    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)
    new_y = Variable("new_y", syntax_settings.DEFAULT_INTEGER)
    stmt = stmt.get_with_updated_variables([(x, new_x), (y, new_y)])
    assert stmt.print_lines() == ["new_x=new_y\n"]


def test_update_variables_binary_operation_node():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    expr = x + y

    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)
    new_y = Variable("new_y", syntax_settings.DEFAULT_INTEGER)
    expr = expr.get_with_updated_variables([(x, new_x), (y, new_y)])
    assert render(expr) == "(new_x+new_y)\n"


def test_update_variables_simple_add():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    expr = x + 5

    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)
    expr = expr.get_with_updated_variables([(x, new_x)])
    assert render(expr) == "(new_x+5_c_int64_t)\n"


def test_update_variables_getattr_node():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    expr = GetAttr(x, "tag")
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)
    expr = expr.get_with_updated_variables([(x, new_x)])
    assert render(expr) == "new_x%tag\n"


def test_update_variables_getitem_node():
    arr = Variable("a", syntax_settings.DEFAULT_REAL, shape=(10, 10))
    expr = arr[1, 2]
    new_arr = Variable("new_a", syntax_settings.DEFAULT_REAL, shape=(40, 30))
    expr = expr.get_with_updated_variables([(arr, new_arr)])
    assert render(expr) == "new_a(1, 2)\n"


def test_update_variables_withgetitem_node():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    expr = Assignment(x, 5, add_to_scope=False)
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER, shape=(10,))
    expr = expr.get_with_updated_variables([(x, new_x[3])])
    assert render(expr) == "new_x(3)=5_c_int64_t\n"


def test_update_variables_eq_ne_nodes():
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)
    new_y = Variable("new_y", syntax_settings.DEFAULT_INTEGER)

    expr = x == y
    expr = expr.get_with_updated_variables([(x, new_x), (y, new_y)])
    assert render(expr) == "(new_x.eq.new_y)\n"

    expr = x != y
    expr = expr.get_with_updated_variables([(x, new_x), (y, new_y)])
    assert render(expr) == "(new_x.ne.new_y)\n"


def test_update_variables_re_im_nodes():
    z = Variable("z", syntax_settings.DEFAULT_COMPLEX)
    new_z = Variable("new_z", syntax_settings.DEFAULT_COMPLEX)

    z_real = z.real
    z_real = z_real.get_with_updated_variables([(z, new_z)])

    z_imag = z.imag
    z_imag = z_imag.get_with_updated_variables([(z, new_z)])

    assert render(z_real) == "new_z%re\n"
    assert render(z_imag) == "new_z%im\n"


@pytest.mark.parametrize(
    "func,nargs,token",
    [
        (Abs, 1, "abs"),
        (Neg, 1, "-"),
        (Not, 1, ".not."),
        (Allocated, 1, "allocated"),
        # TODO(Shape, 1, "shape"),
        (All, 1, "all"),
        (Real, 1, "real"),
        (Imag, 1, "aimag"),
        (Conjugate, 1, "conjg"),
        (Transpose, 1, "transpose"),
        (Exp, 1, "exp"),
        (Sqrt, 1, "sqrt"),
        (Floor, 1, "floor"),
        (Sin, 1, "sin"),
        (Cos, 1, "cos"),
        (Tan, 1, "tan"),
        (Sinh, 1, "sinh"),
        (Cosh, 1, "cosh"),
        (Tanh, 1, "tanh"),
        (ASin, 1, "asin"),
        (ACos, 1, "acos"),
        (ATan, 1, "atan"),
        (Rank, 1, "rank"),
        (Maxval, 1, "maxval"),
        (Minval, 1, "minval"),
        (Popcnt, 1, "popcnt"),
        (Trailz, 1, "trailz"),
        (Sum, 1, "sum"),
        (ATan2, 2, "atan2"),
        (Dotproduct, 2, "dot_product"),
        (Size, 2, "size"),
        (Max, 2, "max"),
        (Min, 2, "min"),
        (Iand, 2, "iand"),
        (Ior, 2, "ior"),
        (Xor, 2, "xor"),
        (Ishft, 2, "ishft"),
        (Ibset, 2, "ibset"),
        (Ibclr, 2, "ibclr"),
        (Matmul, 2, "matmul"),
    ],
)
def test_intrinsic_functions(func, nargs, token):
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    y = Variable("y", syntax_settings.DEFAULT_INTEGER)
    args = [x] if nargs == 1 else [x, y]
    if func is Size:
        args[1] = 1
    expr = func(*args)
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)
    new_y = Variable("new_y", syntax_settings.DEFAULT_INTEGER)
    expr = expr.get_with_updated_variables([(x, new_x), (y, new_y)])

    expected_args = ["new_x"] if nargs == 1 else ["new_x", "new_y"]
    if func is Size:
        expected_args[1] = "1_c_int64_t"
    expected = f"{token}({', '.join(expected_args)})\n"
    assert render(expr) == expected


def test_update_variables_do_statement():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)

    new_i = Variable("new_i", syntax_settings.DEFAULT_INTEGER)
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)

    with Scope():
        do = Do(i, 0, 3, add_to_scope=False)
        with do:
            Assignment(x, i + 1)

        do = do.get_with_updated_variables([(i, new_i), (x, new_x)])

    expected = [
        "do new_i = 0_c_int64_t, 3_c_int64_t\n",
        "    new_x=(new_i+1_c_int64_t)\n",
        "end do\n",
    ]

    for l1, l2 in zip(do.print_lines(), expected):
        assert l1 == l2


def test_update_variables_if_statement():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)

    new_i = Variable("new_i", syntax_settings.DEFAULT_INTEGER)
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)

    with Scope():

        wrapper = Do(i, 0, 3, add_to_scope=False)
        with wrapper:
            with If(i < 5):
                Assignment(x, i + 1)
            with ElseIf(i < 10):
                Assignment(x, i + 2)
            with Else():
                Assignment(x, 0)

        wrapper = wrapper.get_with_updated_variables([(i, new_i), (x, new_x)])

    expected = [
        "do new_i = 0_c_int64_t, 3_c_int64_t\n",
        "    if((new_i.lt.5_c_int64_t))then\n",
        "        new_x=(new_i+1_c_int64_t)\n",
        "    elseif((new_i.lt.10_c_int64_t))then\n",
        "        new_x=(new_i+2_c_int64_t)\n",
        "    else\n",
        "        new_x=0_c_int64_t\n",
        "    end if\n",
        "end do\n",
    ]

    for line1, line2 in zip(wrapper.print_lines(), expected):
        assert line1 == line2, f"Expected: {line2}, but got: {line1}"


def test_update_variables_do_while_statement():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    i = Variable("i", syntax_settings.DEFAULT_INTEGER)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)

    new_i = Variable("new_i", syntax_settings.DEFAULT_INTEGER)
    new_x = Variable("new_x", syntax_settings.DEFAULT_INTEGER)

    with Scope():
        do = DoWhile(i < 5, add_to_scope=False)
        with do:
            Assignment(x, i + 1)

        do = do.get_with_updated_variables([(i, new_i), (x, new_x)])

    expected = [
        "do while ((new_i.lt.5_c_int64_t))\n",
        "    new_x=(new_i+1_c_int64_t)\n",
        "end do\n",
    ]

    for l1, l2 in zip(do.print_lines(), expected):
        assert l1 == l2


def test_call():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    x = Variable("x", syntax_settings.DEFAULT_INTEGER, intent="in")
    y = Variable("y", syntax_settings.DEFAULT_INTEGER, intent="out")
    callee = Subroutine("callee")
    callee.add_variable(x, y)
    with callee.scope:
        Assignment(y, x)
    caller = Subroutine("caller")
    with caller.scope:
        Call(callee, x, y)

    expected = [
        "subroutine caller() bind(C, name='caller')\n",
        "    use iso_c_binding, only: c_int64_t\n",
        "    implicit none\n",
        "    interface\n",
        "        subroutine callee(x, y) bind(C, name='callee')\n",
        "            use iso_c_binding, only: c_int64_t\n",
        "            implicit none\n",
        "            integer(c_int64_t), intent(in), value :: x\n",
        "            integer(c_int64_t), intent(out) :: y\n",
        "        end subroutine callee\n",
        "    end interface\n",
        "    integer(c_int64_t), intent(in), value :: x\n",
        "    integer(c_int64_t), intent(out) :: y\n",
        "    call callee(x, y)\n",
        "end subroutine caller\n",
    ]
    for l1, l2 in zip(caller.print_lines(), expected):
        assert l1 == l2


def test_call_external_module():
    settings.set_default_from_datatype(nm.int64, iso_c=True)
    lib = nm.syntax.module.ExternalModule("module", None, hidden=True)
    lib.add_method("foo", [Variable("a", syntax_settings.DEFAULT_INTEGER)], None)
    foo = lib.foo
    sub = Subroutine("mysub")
    x = Variable("x", syntax_settings.DEFAULT_INTEGER)
    sub.add_variable(x)
    with sub.scope:
        foo(x)

    expected = [
        "subroutine mysub() bind(C, name='mysub')\n",
        "    use iso_c_binding, only: c_int64_t\n",
        "    implicit none\n",
        "    interface\n",
        "        subroutine foo(a)\n",
        "            use iso_c_binding, only: c_int64_t\n",
        "            implicit none\n",
        "            integer(c_int64_t) :: a\n",
        "        end subroutine foo\n",
        "    end interface\n",
        "    integer(c_int64_t) :: x\n",
        "    call foo(x)\n",
        "end subroutine mysub\n",
    ]
    for l1, l2 in zip(sub.print_lines(), expected):
        assert l1 == l2



from numeta.wrappers.external_library import ExternalLibraryWrapper
from numeta.external_modules.iso_c_binding import FPointer_c, FSizet_c


class NumpyMemLib(ExternalLibraryWrapper):
    """Library exposing Numpy allocating and deallocate functions."""

    def __init__(self, directory=None, include=None, additional_flags=None):
        super().__init__("numpy_mem", directory, include, additional_flags, to_link=False)
        self.add_method("numpy_allocate", [FPointer_c, FSizet_c], None)
        self.add_method("numpy_deallocate", [FPointer_c], None)


numpy_mem = NumpyMemLib()



import numpy as np
import numeta as nm


def test_jitted_functions_registry():
    nm.clear_jitted_functions()

    @nm.jit
    def add_one(a):
        a[:] += 1

    @nm.jit
    def add_two(a):
        a[:] += 2

    array = np.zeros(10, dtype=np.int64)
    add_one(array)
    assert all(array == 1)
    add_two(array)
    assert all(array == 3)

    assert len(nm.jitted_functions()) == 2


def test_jitted_functions_registry_clear():
    nm.clear_jitted_functions()

    @nm.jit
    def add_one(a):
        a[:] += 1

    @nm.jit
    def add_two(a):
        a[:] += 2

    array = np.zeros(10, dtype=np.int64)
    add_one(array)
    assert all(array == 1)
    nm.clear_jitted_functions()

    add_two(array)
    assert all(array == 3)

    assert len(nm.jitted_functions()) == 1



import numpy as np
import numeta as nm


def test_disable_array_descriptor():

    nm.settings.unset_add_shape_descriptors()

    @nm.jit
    def callee(n, a):
        # if shape descriptors were not disable i should have done a[0, 0]
        a[0] = n

    a = np.zeros((5, 5), dtype=np.int64)
    callee(3, a)

    expected = np.zeros((5, 5), dtype=np.int64)
    expected[0, 0] = 3
    np.testing.assert_equal(a, expected)

    nm.settings.set_add_shape_descriptors()



from numeta.builder_helper import BuilderHelper
from numeta.syntax import Do
from numeta.settings import syntax_settings


def range(*args):
    if len(args) == 1:
        start = 0
        stop = args[0]
        step = None
    elif len(args) == 2:
        start = args[0]
        stop = args[1]
        step = None
    elif len(args) == 3:
        start = args[0]
        stop = args[1]
        step = args[2]
    else:
        raise ValueError("Invalid number of arguments")

    builder = BuilderHelper.get_current_builder()
    I = builder.generate_local_variables("fc_i", ftype=syntax_settings.DEFAULT_INTEGER)

    with Do(I, start, stop - 1, step=step):
        yield I



from typing import (
    Any,
    Callable,
    Optional,
    overload,
)

from .numeta_function import NumetaFunction, NumetaCompilationTarget
from .registry import clear_registered_functions, registered_functions


def jitted_functions() -> list[NumetaCompilationTarget]:
    """Return a list of all functions compiled via :func:`jit`."""
    return registered_functions()


def clear_jitted_functions() -> None:
    """Clear the registry of jitted functions."""
    clear_registered_functions()


@overload
def jit(func):
    """@jit used with no arguments."""
    ...


@overload
def jit(
    *,
    directory: Optional[str] = None,
    do_checks: bool = True,
    compile_flags: str = "-O3 -march=native",
    namer: Optional[Callable[..., str]] = None,
    inline: bool | int = False,
):
    """@jit(...) used with arguments."""
    ...


def jit(
    func: Callable[..., Any] | None = None,
    *,
    directory: Optional[str] = None,
    do_checks: bool = True,
    compile_flags: str = "-O3 -march=native",
    namer: Optional[Callable[..., str]] = None,
    inline: bool | int = False,
):
    """
    Compile a function with the Numeta JIT, either directly or via parameters.

    Overload Resolution
    -------------------
    1.  **No-arg form**: `@jit`
        - Returns a `NumetaFunction` wrapping the target.
    2.  **With-arg form**: `@jit(directory=..., inline=2, ...)`
        - Returns a decorator that, when applied, produces a `NumetaFunction`.

    Parameters
    ----------
    func
        The function to compile when using `@jit` with no args.
    directory
        Target directory for compiled output (default: none → temp dir).
    do_checks
        Whether to enable compile-time argument validation.
    compile_flags
        Flags for the compiler optimization step.
    namer
        Optional callable to name the JIT-generated symbols.
    inline
        Controls inlining behavior (bool or max-stmts int).

    Returns
    -------
    NumetaFunction
    """
    if func is None:

        def decorator_wrapper(f) -> NumetaFunction:
            return NumetaFunction(
                f,
                directory=directory,
                do_checks=do_checks,
                compile_flags=compile_flags,
                namer=namer,
                inline=inline,
            )

        return decorator_wrapper
    else:
        return NumetaFunction(
            func,
            directory=directory,
            do_checks=do_checks,
            compile_flags=compile_flags,
            namer=namer,
            inline=inline,
        )



from .statement import StatementWithScope
from .various import Comment, Use, Implicit, Interface
from .tools import (
    get_nested_dependencies_or_declarations,
    divide_variables_and_derived_types,
)


class SubroutineDeclaration(StatementWithScope):
    def __init__(self, subroutine):
        self.subroutine = subroutine

        # First check the arguments dependencies
        entities = list(self.subroutine.arguments.values())
        dependencies, declarations = get_nested_dependencies_or_declarations(
            entities, self.subroutine.parent
        )
        (
            self.variables_dec,
            self.derived_types_dec,
            subroutine_decs,
        ) = divide_variables_and_derived_types(declarations)
        self.interfaces = [dec.subroutine for dec in subroutine_decs.values()]

        # Then check the dependencies in the body
        entities = []
        for statement in self.subroutine.scope.get_statements():
            for var in statement.extract_entities():
                if var not in entities:
                    entities.append(var)
        body_dependencies, body_declarations = get_nested_dependencies_or_declarations(
            entities, self.subroutine.parent
        )
        dependencies.update(body_dependencies)

        (
            body_variables_dec,
            body_derived_types_dec,
            body_subroutine_decs,
        ) = divide_variables_and_derived_types(body_declarations)
        self.local_variables = {
            name: dec.variable
            for name, dec in body_variables_dec.items()
            if name not in self.variables_dec
        }
        self.variables_dec.update(body_variables_dec)
        self.derived_types_dec.update(body_derived_types_dec)
        self.interfaces.extend(dec.subroutine for dec in body_subroutine_decs.values())

        self.dependencies = {}
        self.modules_to_import = []

        from numeta.syntax.module import Module

        for dependency, var in dependencies:
            if hasattr(var, "get_interface_declaration"):
                # Should we add the interface?
                # Only if it is not contained in a module, if not the module will take care
                if not isinstance(dependency, Module) or dependency.hidden:
                    self.interfaces.append(var)
            if isinstance(dependency, Module):
                if not dependency.hidden:
                    self.modules_to_import.append((dependency, var))
                if dependency.parent is not None:
                    self.dependencies[dependency.parent.name] = dependency.parent
            else:
                self.dependencies[dependency.name] = dependency

    @property
    def children(self):
        return []

    def extract_entities(self):
        # Assume nothing is visible outside the Subroutine (maybe not okay?)
        yield self.subroutine

    def get_statements(self):
        if self.subroutine.description is not None:
            for line in self.subroutine.description.split("\n"):
                yield Comment(line, add_to_scope=False)

        for dependency, var in self.modules_to_import:
            yield Use(dependency, only=var, add_to_scope=False)

        yield Implicit(implicit_type="none", add_to_scope=False)

        if self.interfaces:
            yield Interface(self.interfaces)

        yield from self.derived_types_dec.values()

        yield from self.variables_dec.values()

        yield from self.subroutine.scope.get_statements()

    def get_start_code_blocks(self):
        result = []

        if self.subroutine.pure:
            result.append("pure")
            result.append(" ")

        if self.subroutine.elemental:
            result.append("elemental")
            result.append(" ")

        result.extend(["subroutine", " ", self.subroutine.name, "("])

        for variable in self.subroutine.arguments.values():
            # TODO to remove dependency on the indent
            if variable.intent is not None:
                result.extend(variable.get_code_blocks())
                result.append(", ")

        if result[-1] == ", ":
            result.pop()
        result.append(")")

        if self.subroutine.bind_c:
            result.extend([" ", f"bind(C, name='{self.subroutine.name}')"])

        return result

    def get_end_code_blocks(self):
        return ["end", " ", "subroutine", " ", self.subroutine.name]


class InterfaceDeclaration(StatementWithScope):
    def __init__(self, subroutine):
        self.subroutine = subroutine

    @property
    def children(self):
        return []

    def extract_entities(self):
        # Assume nothing is visible outside the interface (maybe not okay?)
        yield self.subroutine

    def get_statements(self):
        if self.subroutine.description is not None:
            for line in self.subroutine.description.split("\n"):
                yield Comment(line, add_to_scope=False)

        # First check the arguments dependencies
        entities = list(self.subroutine.arguments.values())
        dependencies, declarations = get_nested_dependencies_or_declarations(
            entities, self.subroutine.parent
        )
        variables_dec, derived_types_dec, _ = divide_variables_and_derived_types(declarations)

        # Now we can construct the subroutine
        for dependency, var in dependencies:
            yield Use(dependency, only=var, add_to_scope=False)

        yield Implicit(implicit_type="none", add_to_scope=False)

        yield from derived_types_dec.values()

        yield from variables_dec.values()

    def get_start_code_blocks(self):
        result = []

        if self.subroutine.pure:
            result.append("pure")
            result.append(" ")

        if self.subroutine.elemental:
            result.append("elemental")
            result.append(" ")

        result.extend(["subroutine", " ", self.subroutine.name, "("])

        for variable in self.subroutine.arguments.values():
            result.extend(variable.get_code_blocks())
            result.append(", ")

        if result[-1] == ", ":
            result.pop()
        result.append(")")

        if self.subroutine.bind_c:
            result.extend([" ", f"bind(C, name='{self.subroutine.name}')"])

        return result

    def get_end_code_blocks(self):
        return ["end", " ", "subroutine", " ", self.subroutine.name]



from .syntax.settings import settings as syntax_settings


class Settings:

    def __init__(
        self,
        iso_C,
        use_numpy_allocator=True,
        reorder_kwargs=True,
        add_shape_descriptors=True,
        ignore_fixed_shape_in_nested_calls=False,
    ):
        """Initialize the settings.
        Parameters
        ----------
        iso_C : bool
            Whether to use ISO C compatibility mode.
        use_numpy_allocator : bool
            Whether to use the NumPy memory allocator.
        reorder_kwargs : bool
            If True, keyword arguments in the generated functions are reordered to ensure
            a unique and deterministic function signature, independent of the order in
            which keywords are passed. This is helpful for reproducibility but may be
            inconvenient when using numeta as a code generator.
        add_shape_descriptors : bool
            If True, shape descriptors are added to array arguments in the generated
            functions. Shape descriptors encode the dimensions of arrays, which is
            typically required when using JIT compilation. However, this can be
            undesirable if numeta is used purely as a code generator.
        ignore_fixed_shape_in_nested_calls : bool
            If True, instead of passing array in a nested call as fixed shape,
            they are passed with undefined dimensions.
            This can help limiting the number of generated functions if they have no dependence on the fixed shape.
        """
        self.iso_C = iso_C
        if self.iso_C:
            self.set_iso_C()
        else:
            self.unset_iso_C()
        self.use_numpy_allocator = use_numpy_allocator
        self.__reorder_kwargs = reorder_kwargs
        self.__add_shape_descriptors = add_shape_descriptors
        self.__ignore_fixed_shape_in_nested_calls = ignore_fixed_shape_in_nested_calls

    def set_default_from_datatype(self, dtype, *, iso_c: bool = False):
        """Set the default Fortran type using a :class:`DataType` subclass."""
        from .datatype import DataType

        if not isinstance(dtype, type) or not issubclass(dtype, DataType):
            raise TypeError("dtype must be a DataType subclass")

        ftype = dtype.get_fortran(bind_c=iso_c)
        syntax_settings.set_default_fortran_type(ftype)

    def set_iso_C(self):
        """Set the ISO C compatibility mode."""
        self.iso_C = True
        syntax_settings.set_c_like()
        from .datatype import (
            int64,
            float64,
            complex128,
            bool8,
            char,
        )

        self.set_default_from_datatype(int64, iso_c=True)
        self.set_default_from_datatype(float64, iso_c=True)
        self.set_default_from_datatype(complex128, iso_c=True)
        self.set_default_from_datatype(bool8, iso_c=True)

        self.set_default_from_datatype(char, iso_c=True)

    def unset_iso_C(self):
        """Unset the ISO C compatibility mode."""
        self.iso_C = False
        syntax_settings.unset_c_like()
        from .datatype import (
            int64,
            float64,
            complex128,
            bool8,
            char,
        )

        self.set_default_from_datatype(int64, iso_c=False)
        self.set_default_from_datatype(float64, iso_c=False)
        self.set_default_from_datatype(complex128, iso_c=False)
        self.set_default_from_datatype(bool8, iso_c=False)
        self.set_default_from_datatype(char, iso_c=False)

    def set_numpy_allocator(self):
        """Set whether to use the NumPy memory allocator."""
        self.use_numpy_allocator = True

    def unset_numpy_allocator(self):
        """Unset the NumPy memory allocator."""
        self.use_numpy_allocator = False

    @property
    def reorder_kwargs(self):
        """Return whether to reorder keyword arguments in the generated function."""
        return self.__reorder_kwargs

    def set_reorder_kwargs(self):
        """Set whether to reorder keyword arguments in the generated function."""
        self.__reorder_kwargs = True

    def unset_reorder_kwargs(self):
        """Unset the reordering of keyword arguments in the generated function."""
        self.__reorder_kwargs = False

    @property
    def add_shape_descriptors(self):
        """Return whether to add shape descriptors to array arguments in generated functions."""
        return self.__add_shape_descriptors

    def set_add_shape_descriptors(self):
        """Set whether to add shape descriptors to array arguments in generated functions."""
        self.__add_shape_descriptors = True

    def unset_add_shape_descriptors(self):
        """Unset the addition of shape descriptors to array arguments in generated functions."""
        self.__add_shape_descriptors = False

    @property
    def ignore_fixed_shape_in_nested_calls(self):
        return self.__ignore_fixed_shape_in_nested_calls

    @ignore_fixed_shape_in_nested_calls.setter
    def ignore_fixed_shape_in_nested_calls(self, value):
        self.__ignore_fixed_shape_in_nested_calls = value


settings = Settings(iso_C=True, use_numpy_allocator=True)
